#!/usr/bin/env python
"""decoder_cells.py: Implementation of RNNSearch in Chainer"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import numpy as np
import chainer
from chainer import cuda, Variable
from chainer import Link, Chain, ChainList
import chainer.functions as F
import chainer.links as L
import random

import rnn_cells
from nmt_chainer.utilities.utils import ortho_init, minibatch_sampling

from nmt_chainer.utilities.constant_batch_mul import batch_matmul_constant, matmul_constant

from attention import AttentionModule

import logging
logging.basicConfig()
log = logging.getLogger("rnns:dec")
log.setLevel(logging.INFO)


class ConditionalizedDecoderCell(object):
    def __init__(self, decoder_chain, compute_ctxt, mb_size, noise_on_prev_word=False,
                 mode="test", lexicon_probability_matrix=None, lex_epsilon=1e-3, demux=False):
        self.decoder_chain = decoder_chain
        self.compute_ctxt = compute_ctxt
        self.noise_on_prev_word = noise_on_prev_word
        self.mode = mode
        self.lexicon_probability_matrix = lexicon_probability_matrix
        self.lex_epsilon = lex_epsilon

        self.mb_size = mb_size
        self.demux = demux

        self.xp = decoder_chain.xp

        if noise_on_prev_word:
            self.noise_mean = self.xp.ones((mb_size, self.decoder_chain.Eo), dtype=self.xp.float32)
            self.noise_lnvar = self.xp.zeros((mb_size, self.decoder_chain.Eo), dtype=self.xp.float32)

    def advance_state(self, previous_states, prev_y):
        current_mb_size = prev_y.data.shape[0]
        assert self.mb_size is None or current_mb_size <= self.mb_size

        if current_mb_size < len(previous_states[0].data):
            truncated_states = [None] * len(previous_states)
            for num_state in xrange(len(previous_states)):
                truncated_states[num_state], _ = F.split_axis(
                    previous_states[num_state], (current_mb_size,), 0)
            previous_states = tuple(truncated_states)

        output_state = previous_states[-1]
        ci, attn = self.compute_ctxt(output_state)
        concatenated = F.concat((prev_y, ci))

        new_states = self.decoder_chain.gru(previous_states, concatenated, mode=self.mode)
        return new_states, concatenated, attn

    def compute_logits(self, new_states, concatenated, attn):
        new_output_state = new_states[-1]

        all_concatenated = F.concat((concatenated, new_output_state))
        logits = self.decoder_chain.lin_o(self.decoder_chain.maxo(all_concatenated))

        if self.lexicon_probability_matrix is not None:
            current_mb_size = new_output_state.data.shape[0]
            assert self.mb_size is None or current_mb_size <= self.mb_size
            lexicon_probability_matrix = self.lexicon_probability_matrix[:current_mb_size]

            # Just making sure data shape is as expected
            attn_mb_size, max_source_length_attn = attn.data.shape
            assert attn_mb_size == current_mb_size
            lex_mb_size, max_source_length_lexicon, v_size_lexicon = lexicon_probability_matrix.shape
            assert max_source_length_lexicon == max_source_length_attn
            assert logits.data.shape == (current_mb_size, v_size_lexicon)

            if self.demux:
                assert lex_mb_size == 1
                weighted_lex_probs = F.reshape(
                    matmul_constant(attn, lexicon_probability_matrix.reshape(lexicon_probability_matrix.shape[1],
                                                                             lexicon_probability_matrix.shape[2])),
                    logits.data.shape)
            else:
                assert lex_mb_size == current_mb_size

    #                 weighted_lex_probs = F.reshape(
    #                         F.batch_matmul(attn, ConstantFunction(lexicon_probability_matrix)(), transa = True),
    #                                                logits.data.shape)

                weighted_lex_probs = F.reshape(
                    batch_matmul_constant(attn, lexicon_probability_matrix, transa=True),
                    logits.data.shape)

            logits += F.log(weighted_lex_probs + self.lex_epsilon)
        return logits

    def advance_one_step(self, previous_states, prev_y):

        if self.noise_on_prev_word:
            current_mb_size = prev_y.data.shape[0]
            assert self.mb_size is None or current_mb_size <= self.mb_size
            prev_y = prev_y * F.gaussian(Variable(self.noise_mean[:current_mb_size], volatile="auto"),
                                         Variable(self.noise_lnvar[:current_mb_size], volatile="auto"))

        new_states, concatenated, attn = self.advance_state(previous_states, prev_y)

        logits = self.compute_logits(new_states, concatenated, attn)

        return new_states, logits, attn

    def get_initial_logits(self, mb_size=None):
        if mb_size is None:
            mb_size = self.mb_size
        assert mb_size is not None

        previous_states = self.decoder_chain.gru.get_initial_states(mb_size)

        prev_y = F.broadcast_to(self.decoder_chain.bos_embeding, (mb_size, self.decoder_chain.Eo))

        new_states, logits, attn = self.advance_one_step(previous_states, prev_y)

        return new_states, logits, attn

    def __call__(self, prev_states, inpt):
        prev_y = self.decoder_chain.emb(inpt)

        new_states, logits, attn = self.advance_one_step(prev_states, prev_y)

        return new_states, logits, attn


def compute_loss_from_decoder_cell(cell, targets, use_previous_prediction=0,
                                   raw_loss_info=False, per_sentence=False, keep_attn=False):
    loss = None
    attn_list = []

    mb_size = targets[0].data.shape[0]
    assert cell.mb_size is None or cell.mb_size == mb_size
    states, logits, attn = cell.get_initial_logits(mb_size)

    total_nb_predictions = 0

    for i in xrange(len(targets)):
        if keep_attn:
            attn_list.append(attn)

        if per_sentence:
            normalized_logits = F.log_softmax(logits)
            total_local_loss = F.select_item(normalized_logits, targets[i])
            if loss is not None and total_local_loss.data.shape[0] != loss.data.shape[0]:
                assert total_local_loss.data.shape[0] < loss.data.shape[0]
                total_local_loss = F.concat(
                    (total_local_loss,
                     Variable(cell.xp.zeros(loss.data.shape[0] - total_local_loss.data.shape[0],
                                            dtype=cell.xp.float32), volatile="auto")),
                    axis=0)
        else:
            local_loss = F.softmax_cross_entropy(logits, targets[i])
            nb_predictions = targets[i].data.shape[0]
            total_local_loss = local_loss * nb_predictions
            total_nb_predictions += nb_predictions

        loss = total_local_loss if loss is None else loss + total_local_loss

        if i >= len(targets) - 1:  # skipping generation of last states as unneccessary
            break

        current_mb_size = targets[i].data.shape[0]
        required_next_mb_size = targets[i + 1].data.shape[0]

        if use_previous_prediction > 0 and random.random() < use_previous_prediction:
            previous_word = Variable(cell.xp.argmax(logits.data[:required_next_mb_size], axis=1).astype(cell.xp.int32), volatile="auto")
        else:
            if required_next_mb_size < current_mb_size:
                previous_word, _ = F.split_axis(targets[i], (required_next_mb_size,), 0)
                current_mb_size = required_next_mb_size
            else:
                previous_word = targets[i]

        states, logits, attn = cell(states, previous_word)

    if raw_loss_info:
        return (loss, total_nb_predictions), attn_list
    else:
        loss = loss / total_nb_predictions
        return loss, attn_list


def sample_from_decoder_cell(cell, nb_steps, best=False, keep_attn_values=False,
                             need_score=False):

    states, logits, attn = cell.get_initial_logits()

    score = 0
    sequences = []
    attn_list = []

    for _ in xrange(nb_steps):
        if keep_attn_values:
            attn_list.append(attn)

        probs = F.softmax(logits)
        if best:
            curr_idx = cell.xp.argmax(probs.data, 1).astype(np.int32)
        else:
            #                 curr_idx = self.xp.empty((mb_size,), dtype = np.int32)
            if cell.xp != np:
                probs_data = cuda.to_cpu(probs.data)
            else:
                probs_data = probs.data
            curr_idx = minibatch_sampling(probs_data)
            if cell.xp != np:
                curr_idx = cuda.to_gpu(curr_idx.astype(np.int32))
            else:
                curr_idx = curr_idx.astype(np.int32)
#                 for i in xrange(mb_size):
#                     sampler = chainer.utils.WalkerAlias(probs_data[i])
#                     curr_idx[i] =  sampler.sample(1)[0]
        if need_score:
            score = score + np.log(cuda.to_cpu(probs.data)[np.arange(cell.mb_size), cuda.to_cpu(curr_idx)])
        sequences.append(curr_idx)

        previous_word = Variable(curr_idx, volatile="auto")

        states, logits, attn = cell(states, previous_word)

    return sequences, score, attn_list


class Decoder(Chain):
    """ Decoder for RNNSearch.
        The __call_ takes 3 required parameters: fb_concat, targets, mask.

        fb_concat should be the result of a call to Encoder.

        targets is a python list of chainer variables of type int32 and of variable shape (n,)
            the values n should be decreasing:
                i < j => targets[i].data.shape[0] >= targets[j].data.shape[0]
            targets[i].data[j] is the jth elements of the ith sequence in the minibatch
            all this imply that the sequences of the minibatch should be sorted from longest to shortest

        mask is as in the description of Encoder.

        * it is up to the user to add an EOS token to the data.

        Return a loss and the attention model values
    """

    def __init__(self, Vo, Eo, Ho, Ha, Hi, Hl, attn_cls=AttentionModule, init_orth=False,
                 cell_type=rnn_cells.LSTMCell):
        #         assert cell_type in "gru dgru lstm slow_gru".split()
        #         self.cell_type = cell_type
        #         if cell_type == "gru":
        #             gru = faster_gru.GRU(Ho, Eo + Hi)
        #         elif cell_type == "dgru":
        #             gru = DoubleGRU(Ho, Eo + Hi)
        #         elif cell_type == "lstm":
        #             gru = L.StatelessLSTM(Eo + Hi, Ho) #, forget_bias_init = 3)
        #         elif cell_type == "slow_gru":
        #             gru = L.GRU(Ho, Eo + Hi)

        if isinstance(cell_type, (str, unicode)):
            cell_type = rnn_cells.create_cell_model_from_string(cell_type)

        gru = cell_type(Eo + Hi, Ho)

        log.info("constructing decoder [%r]" % (cell_type,))

        super(Decoder, self).__init__(
            emb=L.EmbedID(Vo, Eo),
            #             gru = L.GRU(Ho, Eo + Hi),

            gru=gru,

            maxo=L.Maxout(Eo + Hi + Ho, Hl, 2),
            lin_o=L.Linear(Hl, Vo, nobias=False),

            attn_module=attn_cls(Hi, Ha, Ho, init_orth=init_orth)
        )
#         self.add_param("initial_state", (1, Ho))
        self.add_param("bos_embeding", (1, Eo))

        self.Hi = Hi
        self.Ho = Ho
        self.Eo = Eo
#         self.initial_state.data[...] = np.random.randn(Ho)
        self.bos_embeding.data[...] = np.random.randn(Eo)

        if init_orth:
            ortho_init(self.gru)
            ortho_init(self.lin_o)
            ortho_init(self.maxo)

    def give_conditionalized_cell(self, fb_concat, src_mask, noise_on_prev_word=False,
                                  mode="test", lexicon_probability_matrix=None, lex_epsilon=1e-3, demux=False):
        assert mode in "test train".split()
        mb_size, nb_elems, Hi = fb_concat.data.shape
        assert Hi == self.Hi, "%i != %i" % (Hi, self.Hi)

        compute_ctxt = self.attn_module(fb_concat, src_mask)

        if not demux:
            return ConditionalizedDecoderCell(self, compute_ctxt, mb_size, noise_on_prev_word=noise_on_prev_word,
                                              mode=mode, lexicon_probability_matrix=lexicon_probability_matrix, lex_epsilon=lex_epsilon)
        else:
            assert mb_size == 1
            assert demux >= 1
            compute_ctxt = self.attn_module.compute_ctxt_demux(fb_concat, src_mask)
            return ConditionalizedDecoderCell(self, compute_ctxt, None, noise_on_prev_word=noise_on_prev_word,
                                              mode=mode, lexicon_probability_matrix=lexicon_probability_matrix, lex_epsilon=lex_epsilon,
                                              demux=True)

    def compute_loss(self, fb_concat, src_mask, targets, raw_loss_info=False, keep_attn_values=False,
                     noise_on_prev_word=False, use_previous_prediction=0, mode="test", per_sentence=False,
                     lexicon_probability_matrix=None, lex_epsilon=1e-3
                     ):
        decoding_cell = self.give_conditionalized_cell(fb_concat, src_mask, noise_on_prev_word=noise_on_prev_word,
                                                       mode=mode, lexicon_probability_matrix=lexicon_probability_matrix, lex_epsilon=lex_epsilon)
        loss, attn_list = compute_loss_from_decoder_cell(decoding_cell, targets,
                                                         use_previous_prediction=use_previous_prediction,
                                                         raw_loss_info=raw_loss_info,
                                                         per_sentence=per_sentence,
                                                         keep_attn=keep_attn_values)
        return loss, attn_list

    def sample(self, fb_concat, src_mask, nb_steps, mb_size, lexicon_probability_matrix=None,
               lex_epsilon=1e-3, best=False, keep_attn_values=False, need_score=False):
        decoding_cell = self.give_conditionalized_cell(fb_concat, src_mask, noise_on_prev_word=False,
                                                       mode="test", lexicon_probability_matrix=lexicon_probability_matrix,
                                                       lex_epsilon=lex_epsilon)
        sequences, score, attn_list = sample_from_decoder_cell(decoding_cell, nb_steps, best=best,
                                                               keep_attn_values=keep_attn_values,
                                                               need_score=need_score)

        return sequences, score, attn_list



"""server_arg_parsing.py: parse evaluation argument and create configuration dictionnary for server"""

from nmt_chainer.utilities import argument_parsing_tools

_CONFIG_SECTION_TO_DESCRIPTION = {"output": "Output Options",
                                  "process": "Translation Process Options"}


def define_parser(parser):
    parser.add_argument("training_config", nargs="?", help="prefix of the trained model",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    parser.add_argument("trained_model", nargs="?", help="prefix of the trained model",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    parser.add_argument("--host", help="host for listening request", default='localhost')
    parser.add_argument("--port", help="port for listening request", default=44666)
    parser.add_argument("--segmenter_command", help="command to communicate with the segmenter server")
    parser.add_argument("--segmenter_format", help="format to expect from the segmenter (parse_server, morph)", default='plain')

    output_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["output"])
    output_group.add_argument("--tgt_fn", help="target text")
    output_group.add_argument("--nbest_to_rescore", help="nbest list in moses format")
    output_group.add_argument("--ref", help="target text")
    output_group.add_argument("--tgt_unk_id", choices=["attn", "id"], default="align")
    # arguments for unk replace
    output_group.add_argument("--dic")

    management_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["process"])
    management_group.add_argument("--gpu", type=int, help="specify gpu number to use, if any")
    management_group.add_argument("--additional_training_config", nargs="*", help="prefix of the trained model")
    management_group.add_argument("--additional_trained_model", nargs="*", help="prefix of the trained model")
    management_group.add_argument("--max_nb_ex", type=int, help="only use the first MAX_NB_EX examples")
    management_group.add_argument("--mb_size", type=int, default=80, help="Minibatch size")
    management_group.add_argument("--nb_batch_to_sort", type=int, default=20, help="Sort this many batches by size.")
    management_group.add_argument("--reverse_training_config", help="prefix of the trained model")
    management_group.add_argument("--reverse_trained_model", help="prefix of the trained model")


def do_start_server(args):
    import nmt_chainer.translation.server
    nmt_chainer.translation.server.do_start_server(args)


def get_parse_option_orderer():
    description_to_config_section = dict((v, k) for (k, v) in _CONFIG_SECTION_TO_DESCRIPTION.iteritems())
    por = argument_parsing_tools.ParseOptionRecorder(group_title_to_section=description_to_config_section)
    define_parser(por)
    return por


def make_config_server(args):
    parse_option_orderer = get_parse_option_orderer()
    config_eval = parse_option_orderer.convert_args_to_ordered_dict(args)
    config_eval.add_metadata_infos(version_num=1)
    config_eval.set_readonly()

    return config_eval



from nmt_chainer.utilities import argument_parsing_tools
import logging

logging.basicConfig()
log = logging.getLogger("rnns:make_data_config")
log.setLevel(logging.INFO)

_CONFIG_SECTION_TO_DESCRIPTION = {"data": "Files used for training, dev, testing, ...",
                                  "processing": "Processing applied to the data."}


def define_parser(parser):
    parser.add_argument(
        "src_fn", nargs="?",
        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone, help="source language text file for training data")
    parser.add_argument(
        "tgt_fn", nargs="?",
        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone, help="target language text file for training data")
    parser.add_argument(
        "save_prefix", nargs="?",
        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone, help="created files will be saved with this prefix")

    parser.add_argument("--description", help="Optional message to be stored in the configuration file")

    data_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["data"])

    data_group.add_argument("--src_fn", dest="src_fn",
                            action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                            help="same as positional argument src_fn")

    data_group.add_argument("--tgt_fn", dest="tgt_fn",
                            action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                            help="same as positional argument tgt_fn")

    data_group.add_argument("--save_prefix", dest="save_prefix",
                            action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                            help="same as positional argument save_prefix")

    data_group.add_argument("--max_nb_ex", type=int, help="only use the first MAX_NB_EX examples")

    data_group.add_argument("--dev_src", help="specify a source dev set")
    data_group.add_argument("--dev_tgt", help="specify a target dev set")

    data_group.add_argument("--test_src", help="specify a source test set")
    data_group.add_argument("--test_tgt", help="specify a target test set")

    data_group.add_argument("--mode_align", choices=["unk_align", "all_align"])
    data_group.add_argument("--align_fn", help="align file for training data")

    processing_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["processing"])
    processing_group.add_argument("--src_voc_size", type=int, default=32000,
                                  help="limit source vocabulary size to the n most frequent words")
    processing_group.add_argument("--tgt_voc_size", type=int, default=32000,
                                  help="limit target vocabulary size to the n most frequent words")
    processing_group.add_argument("--use_voc", help="specify an exisiting vocabulary file")
#     parser.add_argument("--add_to_valid_set_every", type = int)
#     parser.add_argument("--shuffle", default = False, action = "store_true")
#     parser.add_argument("--enable_fast_shuffle", default = False, action = "store_true")
    processing_group.add_argument("--tgt_segmentation_type", choices=["word", "word2char", "char"], default="word")
    processing_group.add_argument("--src_segmentation_type", choices=["word", "word2char", "char"], default="word")

    processing_group.add_argument("--bpe_src", type=int, help="Apply BPE to source side, with this many merges")
    processing_group.add_argument("--bpe_tgt", type=int, help="Apply BPE to target side, with this many merges")

    processing_group.add_argument("--joint_bpe", type=int, help="Apply joint BPE, with this many merges")

    processing_group.add_argument("--latin_src", default=False, action="store_true", help="apply preprocessing for latin scripts to source")
    processing_group.add_argument("--latin_tgt", default=False, action="store_true", help="apply preprocessing for latin scripts to target")

    processing_group.add_argument("--latin_type", choices="all_adjoint caps_isolate".split(), default="all_adjoint", help="choose preprocessing for latin scripts to source")


class CommandLineValuesException(Exception):
    pass


def cmdline(arguments=None):
    import sys
    import argparse
    parser = argparse.ArgumentParser(description="Prepare training data.",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    define_parser(parser)

    args = parser.parse_args(args=arguments)

    do_make_data(args)

# def get_parse_option_orderer():
#     por = argument_parsing_tools.ParseOptionRecorder()
#     define_parser(por)
#     return por


def get_parse_option_orderer():
    description_to_config_section = dict((v, k) for (k, v) in _CONFIG_SECTION_TO_DESCRIPTION.iteritems())
    por = argument_parsing_tools.ParseOptionRecorder(group_title_to_section=description_to_config_section,
                                                     ignore_positional_arguments=set(["save_prefix", "data_prefix"]))
    define_parser(por)
    return por


def make_data_config(args):
    parse_option_orderer = get_parse_option_orderer()
    config = parse_option_orderer.convert_args_to_ordered_dict(args)
    config.add_metadata_infos(1.0, overwrite=True)
    config.set_readonly()

    if config["data"]["src_fn"] is None or config["data"]["tgt_fn"] is None or config["data"]["save_prefix"] is None:
        raise CommandLineValuesException("src_fn, tgt_fn and save_prefix need to be set either on the command line or in a config file")

    if not ((config.data.test_src is None) == (config.data.test_tgt is None)):
        raise CommandLineValuesException(
            "Command Line Error: either specify both --test_src and --test_tgt or neither")

    if not ((config.data.dev_src is None) == (config.data.dev_tgt is None)):
        raise CommandLineValuesException("Command Line Error: either specify both --dev_src and --dev_tgt or neither")

    if config.processing.joint_bpe is not None and (config.processing.bpe_src is not None or config.processing.bpe_tgt is not None):
        raise CommandLineValuesException("Command Line Error: --joint_bpe is incompatible with --bpe_src and --bpe_tgt")
    return config


def load_config(filename):
    config = argument_parsing_tools.OrderedNamespace.load_from(filename)
    if "metadata" not in config:  # older config file
        parse_option_orderer = get_parse_option_orderer()
        config = parse_option_orderer.convert_args_to_ordered_dict(config, args_is_namespace=False)
    config.set_readonly()
    return config


def do_make_data(args):
    import nmt_chainer.dataprocessing.make_data
    config = make_data_config(args)
    nmt_chainer.dataprocessing.make_data.do_make_data(config)


if __name__ == '__main__':
    cmdline()



#!/usr/bin/env python
"""eval.py: Use a RNNSearch Model"""

__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

from nmt_chainer.utilities.utils import make_batch_src, make_batch_src_tgt, minibatch_provider, compute_bleu_with_unk_as_wrong, de_batch
import logging
import numpy as np
import math
import codecs
import operator
import beam_search
# import h5py

logging.basicConfig()
log = logging.getLogger("rnns:evaluation")
log.setLevel(logging.INFO)


def translate_to_file(encdec, eos_idx, test_src_data, mb_size, tgt_indexer,
                      translations_fn, test_references=None, control_src_fn=None, src_indexer=None, gpu=None, nb_steps=50,
                      reverse_src=False, reverse_tgt=False,
                      s_unk_tag="#S_UNK#", t_unk_tag="#T_UNK#"):

    log.info("computing translations")
    translations = greedy_batch_translate(encdec, eos_idx, test_src_data,
                                          batch_size=mb_size, gpu=gpu, nb_steps=nb_steps,
                                          reverse_src=reverse_src, reverse_tgt=reverse_tgt)

    log.info("writing translation of set to %s" % translations_fn)
    out = codecs.open(translations_fn, "w", encoding="utf8")
    for t in translations:
        if t[-1] == eos_idx:
            t = t[:-1]
#         out.write(convert_idx_to_string(t, tgt_voc) + "\n")
        out.write(tgt_indexer.deconvert(t, unk_tag=t_unk_tag) + "\n")

    if control_src_fn is not None:
        assert src_indexer is not None
        control_out = codecs.open(control_src_fn, "w", encoding="utf8")
        log.info("writing src of test set to %s" % control_src_fn)
        for s in test_src_data:
            #             control_out.write(convert_idx_to_string(s, src_voc) + "\n")
            control_out.write(src_indexer.deconvert(s, unk_tag=s_unk_tag) + "\n")

    if test_references is not None:
        #         unk_id = tgt_indexer.get_unk_idx()  #len(tgt_voc) - 1
        new_unk_id_ref = len(tgt_indexer) + 7777
        new_unk_id_cand = len(tgt_indexer) + 9999
        bc = compute_bleu_with_unk_as_wrong(test_references, [t[:-1] for t in translations], tgt_indexer.is_unk_idx, new_unk_id_ref, new_unk_id_cand)
        log.info("bleu: %r" % bc)
        return bc
    else:
        return None


def compute_loss_all(encdec, test_data, eos_idx, mb_size, gpu=None, reverse_src=False, reverse_tgt=False):
    mb_provider_test = minibatch_provider(test_data, eos_idx, mb_size, nb_mb_for_sorting=-1, loop=False,
                                          gpu=gpu, volatile="on",
                                          reverse_src=reverse_src, reverse_tgt=reverse_tgt)
    test_loss = 0
    test_nb_predictions = 0
    for src_batch, tgt_batch, src_mask in mb_provider_test:
        loss, attn = encdec(src_batch, tgt_batch, src_mask, raw_loss_info=True, mode="test")
        test_loss += loss[0].data
        test_nb_predictions += loss[1]
    test_loss /= test_nb_predictions
    return test_loss


def greedy_batch_translate(encdec, eos_idx, src_data, batch_size=80, gpu=None, get_attention=False, nb_steps=50,
                           reverse_src=False, reverse_tgt=False):
    nb_ex = len(src_data)
    nb_batch = nb_ex / batch_size + (1 if nb_ex % batch_size != 0 else 0)
    res = []
    attn_all = []
    for i in range(nb_batch):
        current_batch_raw_data = src_data[i * batch_size: (i + 1) * batch_size]

        if reverse_src:
            current_batch_raw_data_new = []
            for src_side in current_batch_raw_data:
                current_batch_raw_data_new.append(src_side[::-1])
            current_batch_raw_data = current_batch_raw_data_new

        src_batch, src_mask = make_batch_src(current_batch_raw_data, gpu=gpu, volatile="on")
        sample_greedy, score, attn_list = encdec(src_batch, nb_steps, src_mask, use_best_for_sample=True,
                                                 keep_attn_values=get_attention, mode="test")
        deb = de_batch(sample_greedy, mask=None, eos_idx=eos_idx, is_variable=False)
        res += deb
        if get_attention:
            deb_attn = de_batch(attn_list, mask=None, eos_idx=None, is_variable=True, raw=True,
                                reverse=reverse_tgt)
            attn_all += deb_attn

    if reverse_tgt:
        new_res = []
        for t in res:
            if t[-1] == eos_idx:
                new_res.append(t[:-1][::-1] + [t[-1]])
            else:
                new_res.append(t[::-1])

        res = new_res

    if get_attention:
        assert not reverse_tgt, "not implemented"
        return res, attn_all
    else:
        return res


def reverse_rescore(encdec, src_batch, src_mask, eos_idx, translations, gpu=None):
    from nmt_chainer.utilities import utils

    reversed_translations = []
    for t in translations:
        if t[-1] == eos_idx:
            t = t[:-1]
        reversed_translations.append(t[::-1])

    scorer = encdec.nbest_scorer(src_batch, src_mask)
    tgt_batch, arg_sort = utils.make_batch_tgt(reversed_translations,
                                               eos_idx=eos_idx, gpu=gpu, volatile="on", need_arg_sort=True)

    scores, attn = scorer(tgt_batch)
    scores, _ = scores
    scores = scores.data

    assert len(arg_sort) == len(scores)
    de_sorted_scores = [None] * len(scores)
    for xpos in xrange(len(arg_sort)):
        original_pos = arg_sort[xpos]
        de_sorted_scores[original_pos] = scores[xpos]
    return de_sorted_scores


def beam_search_translate(encdec, eos_idx, src_data, beam_width=20, beam_pruning_margin=None, nb_steps=50, gpu=None,
                          beam_score_coverage_penalty=None, beam_score_coverage_penalty_strength=0.2,
                          need_attention=False, nb_steps_ratio=None, beam_score_length_normalization='none', beam_score_length_normalization_strength=0.2, post_score_length_normalization='simple', post_score_length_normalization_strength=0.2,
                          post_score_coverage_penalty='none', post_score_coverage_penalty_strength=0.2,
                          groundhog=False, force_finish=False,
                          prob_space_combination=False,
                          reverse_encdec=None, use_unfinished_translation_if_none_found=False):
    nb_ex = len(src_data)
#     res = []
    for num_ex in range(nb_ex):
        src_batch, src_mask = make_batch_src([src_data[num_ex]], gpu=gpu, volatile="on")
        assert len(src_mask) == 0
        if nb_steps_ratio is not None:
            nb_steps = int(len(src_data[num_ex]) * nb_steps_ratio) + 1

#         if isinstance(encdec, (tuple, list)):
#             assert len(encdec) == 1
#             encdec = encdec[0]
#
#         translations = encdec.beam_search(src_batch, src_mask, nb_steps = nb_steps, eos_idx = eos_idx,
#                                           beam_width = beam_width,
#                                           beam_opt = beam_opt, need_attention = need_attention,
#                                     groundhog = groundhog)

        if not isinstance(encdec, (tuple, list)):
            encdec = [encdec]
        translations = beam_search.ensemble_beam_search(encdec, src_batch, src_mask, nb_steps=nb_steps, eos_idx=eos_idx,
                                                        beam_width=beam_width,
                                                        beam_pruning_margin=beam_pruning_margin,
                                                        beam_score_length_normalization=beam_score_length_normalization,
                                                        beam_score_length_normalization_strength=beam_score_length_normalization_strength,
                                                        beam_score_coverage_penalty=beam_score_coverage_penalty,
                                                        beam_score_coverage_penalty_strength=beam_score_coverage_penalty_strength,
                                                        need_attention=need_attention, force_finish=force_finish,
                                                        prob_space_combination=prob_space_combination,
                                                        use_unfinished_translation_if_none_found=use_unfinished_translation_if_none_found)

        # TODO: This is a quick patch, but actually ensemble_beam_search probably should not return empty translations except when no translation found
        if len(translations) > 1:
            translations = [t for t in translations if len(t[0]) > 0]

#         print "nb_trans", len(translations), [score for _, score in translations]
#         bests = []
#         translations.sort(key = itemgetter(1), reverse = True)
#         bests.append(translations[0])

        if reverse_encdec is not None and len(translations) > 1:
            rescored_translations = []
            reverse_scores = reverse_rescore(
                reverse_encdec, src_batch, src_mask, eos_idx, [
                    t[0] for t in translations], gpu)
            for num_t in xrange(len(translations)):
                tr, sc, attn = translations[num_t]
                rescored_translations.append(
                    (tr, sc + reverse_scores[num_t], attn))
            translations = rescored_translations

        xp = encdec[0].xp

        if post_score_length_normalization == 'none' and post_score_coverage_penalty == 'none':
            ranking_criterion = operator.itemgetter(1)
        else:
            def ranking_criterion(x):
                length_normalization = 1
                if post_score_length_normalization == 'simple':
                    length_normalization = len(x[0]) + 1
                elif post_score_length_normalization == 'google':
                    length_normalization = pow((len(x[0]) + 5), post_score_length_normalization_strength) / pow(6, post_score_length_normalization_strength)

                coverage_penalty = 0
                if post_score_coverage_penalty == 'google':
                    assert len(src_data[num_ex]) == x[2][0].shape[0]

                    # log.info("sum={0}".format(sum(x[2])))
                    # log.info("min={0}".format(xp.minimum(sum(x[2]), xp.array(1.0))))
                    # log.info("log={0}".format(xp.log(xp.minimum(sum(x[2]), xp.array(1.0)))))
                    log_of_min_of_sum_over_j = xp.log(xp.minimum(sum(x[2]), xp.array(1.0)))
                    coverage_penalty = post_score_coverage_penalty_strength * xp.sum(log_of_min_of_sum_over_j)
                    # log.info("cp={0}".format(coverage_penalty))
                    # cp = 0
                    # for i in xrange(len(src_data[num_ex])):
                    #    attn_sum = 0
                    #    for j in xrange(len(x[0])):
                    #        attn_sum += x[2][j][i]
                    #    #log.info("attn_sum={0}".format(attn_sum))
                    #    #log.info("min={0}".format(min(attn_sum, 1.0)))
                    #    #log.info("log={0}".format(math.log(min(attn_sum, 1.0))))
                    #    cp += math.log(min(attn_sum, 1.0))
                    # log.info("cp={0}".format(cp))
                    # cp *= post_score_coverage_penalty_strength

                    # slow = x[1]/length_normalization + cp
                    # opti = x[1]/length_normalization + coverage_penalty
                    # log.info("type={0}....{1}".format(type(slow), type(opti)))
                    # log.info("shape={0} size={1} dim={2} data={3} elem={4}".format(opti.shape, opti.size, opti.ndim, opti.data, opti.item(0)))
                    # test = '!!!'
                    # if "{0}".format(slow) == "{0}".format(opti):
                    #    test = ''
                    # log.info("score slow <=> optimized: {0} <=> {1} {2}".format(slow, opti, test))

                return x[1] / length_normalization + coverage_penalty

        translations.sort(key=ranking_criterion, reverse=True)

#         bests.append(translations[0])
#         yield bests
        yield translations[0]
#         res.append(bests)
#     return res


def batch_align(encdec, eos_idx, src_tgt_data, batch_size=80, gpu=None):
    nb_ex = len(src_tgt_data)
    nb_batch = nb_ex / batch_size + (1 if nb_ex % batch_size != 0 else 0)
    sum_loss = 0
    attn_all = []
    for i in range(nb_batch):
        current_batch_raw_data = src_tgt_data[i * batch_size: (i + 1) * batch_size]
#         print current_batch_raw_data
        src_batch, tgt_batch, src_mask, arg_sort = make_batch_src_tgt(
            current_batch_raw_data, eos_idx=eos_idx, gpu=gpu, volatile="on", need_arg_sort=True)
        loss, attn_list = encdec(src_batch, tgt_batch, src_mask, keep_attn_values=True)
        deb_attn = de_batch(attn_list, mask=None, eos_idx=None, is_variable=True, raw=True)

        assert len(arg_sort) == len(deb_attn)
        de_sorted_attn = [None] * len(deb_attn)
        for xpos in xrange(len(arg_sort)):
            original_pos = arg_sort[xpos]
            de_sorted_attn[original_pos] = deb_attn[xpos]

        attn_all += de_sorted_attn
        sum_loss += float(loss.data)
    return sum_loss, attn_all

# def convert_idx_to_string(seq, voc, eos_idx = None):
#     trans = []
#     for idx_tgt in seq:
#         if eos_idx is not None and idx_tgt == eos_idx:
#             trans.append("#EOS#")
#         else:
#             if idx_tgt >= len(voc):
#                 log.warn("found unknown idx in tgt : %i / %i"% (idx_tgt, len(voc)))
#             else:
#                 trans.append(voc[idx_tgt])
#     return " ".join(trans)
#
# def convert_idx_to_string_with_attn(seq, voc, attn, unk_idx, unk_pattern = "#T_UNK_%i#"):
#     trans = []
#     for num, idx_tgt in enumerate(seq):
#         if idx_tgt == unk_idx:
#             a = attn[num]
#             xp = cuda.get_array_module(a)
#             src_pos = int(xp.argmax(a))
#             trans.append(unk_pattern%src_pos)
#         else:
#             if idx_tgt >= len(voc):
#                 log.warn("found unknown idx in tgt : %i / %i"% (idx_tgt, len(voc)))
#             else:
#                 trans.append(voc[idx_tgt])
#     return " ".join(trans)


def sample_once(encdec, src_batch, tgt_batch, src_mask, src_indexer, tgt_indexer, eos_idx, max_nb=None,
                s_unk_tag="#S_UNK#", t_unk_tag="#T_UNK#"):
    print "sample"
    sample_greedy, score, attn_list = encdec(src_batch, 50, src_mask, use_best_for_sample=True, need_score=True,
                                             mode="test")


#                 sample, score = encdec(src_batch, 50, src_mask, use_best_for_sample = False)
    assert len(src_batch[0].data) == len(tgt_batch[0].data)
    assert len(sample_greedy[0]) == len(src_batch[0].data)

    debatched_src = de_batch(src_batch, mask=src_mask, eos_idx=None, is_variable=True)
    debatched_tgt = de_batch(tgt_batch, eos_idx=eos_idx, is_variable=True)
    debatched_sample = de_batch(sample_greedy, eos_idx=eos_idx)

    sample_random, score_random, attn_list_random = encdec(src_batch, 50, src_mask, use_best_for_sample=False, need_score=True,
                                                           mode="test")
    debatched_sample_random = de_batch(sample_random, eos_idx=eos_idx)

    for sent_num in xrange(len(debatched_src)):
        if max_nb is not None and sent_num > max_nb:
            break
        src_idx_seq = debatched_src[sent_num]
        tgt_idx_seq = debatched_tgt[sent_num]
        sample_idx_seq = debatched_sample[sent_num]
        sample_random_idx_seq = debatched_sample_random[sent_num]

        print "sent num", sent_num

        for name, seq, unk_tag, indexer, this_eos_idx in zip("src tgt sample sample_random".split(" "),
                                                             [src_idx_seq, tgt_idx_seq, sample_idx_seq, sample_random_idx_seq],
                                                             [s_unk_tag, t_unk_tag, t_unk_tag, t_unk_tag],
                                                             [src_indexer, tgt_indexer, tgt_indexer, tgt_indexer],
                                                             [None, eos_idx, eos_idx, eos_idx]):
            print name, "idx:", seq
            print name, "raw:", " ".join(indexer.deconvert_swallow(seq, unk_tag=unk_tag, eos_idx=this_eos_idx)).encode('utf-8')
            print name, "postp:", indexer.deconvert(seq, unk_tag=unk_tag, eos_idx=this_eos_idx).encode('utf-8')



#!/usr/bin/env python
"""training.py: training procedures."""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

from chainer import serializers
import time

import logging
import sys
# import h5py

import math

from nmt_chainer.utilities.utils import minibatch_provider, minibatch_provider_curiculum
from nmt_chainer.translation.evaluation import (
    compute_loss_all, translate_to_file, sample_once)

logging.basicConfig()
log = logging.getLogger("rnns:training")
log.setLevel(logging.INFO)


def sent_complexity(sent):
    rank_least_common_word = max(sent)
    length = len(sent)
    return length * math.log(rank_least_common_word + 1)


def example_complexity(ex):
    return sent_complexity(ex[0]) + sent_complexity(ex[1])


def train_on_data(encdec, optimizer, training_data, output_files_dict,
                  src_indexer, tgt_indexer, eos_idx, mb_size=80,
                  nb_of_batch_to_sort=20,
                  test_data=None, dev_data=None, valid_data=None,
                  gpu=None, report_every=200, randomized=False,
                  reverse_src=False, reverse_tgt=False, max_nb_iters=None, do_not_save_data_for_resuming=False,
                  noise_on_prev_word=False, curiculum_training=False,
                  use_previous_prediction=0, no_report_or_save=False,
                  use_memory_optimization=False, sample_every=200,
                  use_reinf=False,
                  save_ckpt_every=2000):
    #     ,
    #                   lexical_probability_dictionary = None,
    #                   V_tgt = None,
    #                   lexicon_prob_epsilon = 1e-3):

    if curiculum_training:
        log.info("Sorting training data by complexity")
        training_data_sorted_by_complexity = sorted(training_data, key=example_complexity)
        log.info("done")

        for s, t in training_data_sorted_by_complexity[:400]:
            print example_complexity((s, t))
            print src_indexer.deconvert(s)
            print tgt_indexer.deconvert(t)
            print

        mb_provider = minibatch_provider_curiculum(training_data_sorted_by_complexity, eos_idx, mb_size, nb_of_batch_to_sort, gpu=gpu,
                                                   randomized=randomized, sort_key=lambda x: len(x[0]),
                                                   reverse_src=reverse_src, reverse_tgt=reverse_tgt)
    else:
        mb_provider = minibatch_provider(training_data, eos_idx, mb_size, nb_of_batch_to_sort, gpu=gpu,
                                         randomized=randomized, sort_key=lambda x: len(x[0]),
                                         reverse_src=reverse_src, reverse_tgt=reverse_tgt)

#     mb_provider = minibatch_provider(training_data, eos_idx, mb_size, nb_of_batch_to_sort, gpu = gpu,
#                                      randomized = randomized, sort_key = lambda x:len(x[1]))

    def s_unk_tag(num, utag):
        return "S_UNK_%i" % utag

    def t_unk_tag(num, utag):
        return "T_UNK_%i" % utag

    def save_model(suffix):
        if suffix == "final":
            fn_save = output_files_dict["model_final"]
        elif suffix == "ckpt":
            fn_save = output_files_dict["model_ckpt"]
        elif suffix == "best":
            fn_save = output_files_dict["model_best"]
        elif suffix == "best_loss":
            fn_save = output_files_dict["model_best_loss"]
        else:
            assert False
        log.info("saving model to %s" % fn_save)
        serializers.save_npz(fn_save, encdec)

    def train_once(src_batch, tgt_batch, src_mask):  # , lexicon_matrix = None):
        t0 = time.clock()
        encdec.zerograds()
        t1 = time.clock()
        (total_loss, total_nb_predictions), attn = encdec(src_batch, tgt_batch, src_mask, raw_loss_info=True,
                                                          noise_on_prev_word=noise_on_prev_word,
                                                          use_previous_prediction=use_previous_prediction,
                                                          mode="train")
#         ,
#                                                           lexicon_probability_matrix = lexicon_matrix,
#                                                           lex_epsilon = lexicon_prob_epsilon)
        loss = total_loss / total_nb_predictions
        t2 = time.clock()
        loss.backward()
        t3 = time.clock()
        optimizer.update()
        t4 = time.clock()
        print "loss:", loss.data,
        print " time %f zgrad:%f fwd:%f bwd:%f upd:%f" % (t4 - t0, t1 - t0, t2 - t1, t3 - t2, t4 - t3)
        return float(total_loss.data), total_nb_predictions

    def train_once_optim(src_batch, tgt_batch, src_mask):
        t0 = time.clock()
        encdec.zerograds()
        t1 = time.clock()
        loss, total_nb_predictions = encdec.compute_loss_and_backward(
            src_batch, tgt_batch, src_mask)
        t2 = time.clock()
        print "loss:", loss,
        t3 = time.clock()
        optimizer.update()
        t4 = time.clock()
        print " time %f zgrad:%f fwd:%f bwd:%f upd:%f" % (t4 - t0, t1 - t0, t2 - t1, t3 - t2, t4 - t3)
        return float(loss) * total_nb_predictions, total_nb_predictions

    def train_once_reinf(src_batch, tgt_batch, src_mask):  # , lexicon_matrix = None):
        t0 = time.clock()
        encdec.zerograds()
        t1 = time.clock()

        from nmt_chainer.utilities import utils
        test_ref = utils.de_batch(tgt_batch, is_variable=True)

        reinf_loss = encdec.get_reinf_loss(src_batch, src_mask, eos_idx,
                                           test_ref, nb_steps=50, nb_samples=5,
                                           use_best_for_sample=False,
                                           temperature=None,
                                           mode="test")

        t2 = time.clock()
        reinf_loss.backward()
        t3 = time.clock()
        optimizer.update()
        t4 = time.clock()
        print "reinf loss:", reinf_loss.data, reinf_loss.data / len(src_batch)
        print " time %f zgrad:%f fwd:%f bwd:%f upd:%f" % (t4 - t0, t1 - t0, t2 - t1, t3 - t2, t4 - t3)
        return float(reinf_loss.data), len(src_batch)

    if test_data is not None:
        test_src_data = [x for x, y in test_data]
        test_references = [y for x, y in test_data]

        def translate_test():
            translations_fn = output_files_dict["test_translation_output"]  # save_prefix + ".test.out"
            control_src_fn = output_files_dict["test_src_output"]  # save_prefix + ".test.src.out"
            return translate_to_file(encdec, eos_idx, test_src_data, mb_size, tgt_indexer,
                                     translations_fn, test_references=test_references, control_src_fn=control_src_fn,
                                     src_indexer=src_indexer, gpu=gpu, nb_steps=50, reverse_src=reverse_src, reverse_tgt=reverse_tgt,
                                     s_unk_tag=s_unk_tag, t_unk_tag=t_unk_tag)

        def compute_test_loss():
            log.info("computing test loss")
            test_loss = compute_loss_all(encdec, test_data, eos_idx, mb_size, gpu=gpu,
                                         reverse_src=reverse_src, reverse_tgt=reverse_tgt)
            log.info("test loss: %f" % test_loss)
            return test_loss
    else:
        def translate_test():
            log.info("translate_test: No test data given")
            return None

        def compute_test_loss():
            log.info("compute_test_loss: No test data given")
            return None

    if dev_data is not None:
        dev_src_data = [x for x, y in dev_data]
        dev_references = [y for x, y in dev_data]

        def translate_dev():
            translations_fn = output_files_dict["dev_translation_output"]  # save_prefix + ".test.out"
            control_src_fn = output_files_dict["dev_src_output"]  # save_prefix + ".test.src.out"
            return translate_to_file(encdec, eos_idx, dev_src_data, mb_size, tgt_indexer,
                                     translations_fn, test_references=dev_references, control_src_fn=control_src_fn,
                                     src_indexer=src_indexer, gpu=gpu, nb_steps=50, reverse_src=reverse_src, reverse_tgt=reverse_tgt,
                                     s_unk_tag=s_unk_tag, t_unk_tag=t_unk_tag)

        def compute_dev_loss():
            log.info("computing dev loss")
            dev_loss = compute_loss_all(encdec, dev_data, eos_idx, mb_size, gpu=gpu,
                                        reverse_src=reverse_src, reverse_tgt=reverse_tgt)
            log.info("dev loss: %f" % dev_loss)
            return dev_loss
    else:
        def translate_dev():
            log.info("translate_dev: No dev data given")
            return None

        def compute_dev_loss():
            log.info("compute_dev_loss: No dev data given")
            return None

    if valid_data is not None:
        valid_src_data = [x for x, y in valid_data]
        valid_references = [y for x, y in valid_data]

        def translate_valid():
            translations_fn = output_files_dict["valid_translation_output"]  # save_prefix + ".test.out"
            control_src_fn = output_files_dict["valid_src_output"]  # save_prefix + ".test.src.out"
            return translate_to_file(encdec, eos_idx, valid_src_data, mb_size, tgt_indexer,
                                     translations_fn, test_references=valid_references, control_src_fn=control_src_fn,
                                     src_indexer=src_indexer, gpu=gpu, nb_steps=50, reverse_src=reverse_src, reverse_tgt=reverse_tgt,
                                     s_unk_tag=s_unk_tag, t_unk_tag=t_unk_tag)

        def compute_valid_loss():
            log.info("computing valid loss")
            dev_loss = compute_loss_all(encdec, valid_data, eos_idx, mb_size, gpu=gpu,
                                        reverse_src=reverse_src, reverse_tgt=reverse_tgt)
            log.info("valid loss: %f" % dev_loss)
            return dev_loss
    else:
        def translate_valid():
            log.info("translate_valid: No valid data given")
            return None

        def compute_valid_loss():
            log.info("compute_valid_loss: No valid data given")
            return None

    try:
        best_dev_bleu = 0
        best_dev_loss = None
        prev_time = time.clock()
        prev_i = None
        total_loss_this_interval = 0
        total_nb_predictions_this_interval = 0
        for i in xrange(sys.maxsize):
            if max_nb_iters is not None and max_nb_iters <= i:
                break
            print i,
            src_batch, tgt_batch, src_mask = mb_provider.next()
            if src_batch[0].data.shape[0] != mb_size:
                log.warn("got minibatch of size %i instead of %i" % (src_batch[0].data.shape[0], mb_size))

#             if lexical_probability_dictionary is not None:
#                 lexicon_matrix = utils.compute_lexicon_matrix(src_batch, lexical_probability_dictionary)
#                 if gpu is not None:
#                     lexicon_matrix = cuda.to_gpu(lexicon_matrix, gpu)
#             else:
#                 lexicon_matrix = None

#             if i%100 == 0:
#                 print "valid",
#                 compute_valid()
            if not no_report_or_save:
                if i % sample_every == 0:
                    for v in src_batch + tgt_batch:
                        v.volatile = "on"
                    sample_once(encdec, src_batch, tgt_batch, src_mask, src_indexer, tgt_indexer, eos_idx,
                                max_nb=20,
                                s_unk_tag=s_unk_tag, t_unk_tag=t_unk_tag)
                    for v in src_batch + tgt_batch:
                        v.volatile = "off"
                if i % report_every == 0:
                    current_time = time.clock()
                    if prev_i is not None:
                        iteration_interval = i - prev_i
                        avg_time = (current_time - prev_time) / (iteration_interval)
                        avg_training_loss = total_loss_this_interval / total_nb_predictions_this_interval
                        avg_sentence_size = float(total_nb_predictions_this_interval) / (iteration_interval * mb_size)

                    else:
                        avg_time = 0
                        avg_training_loss = 0
                        avg_sentence_size = 0
                    prev_i = i
                    total_loss_this_interval = 0
                    total_nb_predictions_this_interval = 0

                    print "avg time:", avg_time
                    print "avg training loss:", avg_training_loss
                    print "avg sentence size", avg_sentence_size

                    bc_test = translate_test()
                    test_loss = compute_test_loss()
                    bc_dev = translate_dev()
                    dev_loss = compute_dev_loss()
                    bc_valid = translate_valid()
                    valid_loss = compute_valid_loss()

                    if dev_loss is not None and (best_dev_loss is None or dev_loss <= best_dev_loss):
                        best_dev_loss = dev_loss
                        log.info("saving best loss model %f" % best_dev_loss)
                        save_model("best_loss")

                    if bc_test is not None:

                        assert test_loss is not None
                        import sqlite3
                        import datetime
                        db_path = output_files_dict["sqlite_db"]
                        log.info("saving test results to %s" % (db_path))
                        db_connection = sqlite3.connect(db_path)
                        db_cursor = db_connection.cursor()
                        db_cursor.execute('''CREATE TABLE IF NOT EXISTS exp_data
        (date text, bleu_info text, iteration real,
        loss real, bleu real,
        dev_loss real, dev_bleu real,
        valid_loss real, valid_bleu real,
        avg_time real, avg_training_loss real)''')
                        infos = (datetime.datetime.now().strftime("%I:%M%p %B %d, %Y"),
                                 repr(bc_test), i, float(test_loss), bc_test.bleu(),
                                 float(dev_loss), bc_dev.bleu(),
                                 float(valid_loss) if valid_loss is not None else None, bc_valid.bleu() if bc_valid is not None else None,
                                 avg_time, avg_training_loss)
                        db_cursor.execute("INSERT INTO exp_data VALUES (?,?,?,?,?,?,?,?,?,?,?)", infos)
                        db_connection.commit()
                        db_connection.close()

                        if bc_dev.bleu() > best_dev_bleu:
                            best_dev_bleu = bc_dev.bleu()
                            log.info("saving best model %f" % best_dev_bleu)
                            save_model("best")
                    prev_time = time.clock()
                if i % save_ckpt_every == 0:
                    save_model("ckpt")
                    fn_save_optimizer = output_files_dict["optimizer_ckpt"]
                    log.info(
                        "saving optimizer parameters to %s" %
                        fn_save_optimizer)
                    serializers.save_npz(fn_save_optimizer, optimizer)

            if use_memory_optimization:
                #                 if lexicon_matrix is not None:
                #                     raise NotImplemented
                total_loss, total_nb_predictions = train_once_optim(src_batch, tgt_batch, src_mask)
            elif use_reinf:
                total_loss, total_nb_predictions = train_once_reinf(src_batch, tgt_batch, src_mask)
            else:
                total_loss, total_nb_predictions = train_once(src_batch, tgt_batch, src_mask)
#                 ,
#                                                               lexicon_matrix = lexicon_matrix)

            total_loss_this_interval += total_loss
            total_nb_predictions_this_interval += total_nb_predictions
    finally:
        if not do_not_save_data_for_resuming and not no_report_or_save:
            save_model("final")
            fn_save_optimizer = output_files_dict["optimizer_final"]
            log.info("saving optimizer parameters to %s" % fn_save_optimizer)
            serializers.save_npz(fn_save_optimizer, optimizer)



import argparse
from nmt_chainer.utilities import graph_training
from nmt_chainer.utilities import replace_tgt_unk
from nmt_chainer.utilities import expe_recap
from nmt_chainer.utilities import bleu_computer


def define_parser(parser):
    subparsers = parser.add_subparsers(dest="__sub_subcommand_name")

    graph_parser = subparsers.add_parser('graph', description="Create a graph of the training data.",
                                         help="Create a graph of the training data.", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    graph_training.define_parser(graph_parser)

    replace_tgt_parser = subparsers.add_parser('replace_tgt_unk', description="Replace UNK tags using a dictionnary.",
                                               help="Replace UNK tags using a dictionnary.", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    replace_tgt_unk.define_parser(replace_tgt_parser)

    recap_parser = subparsers.add_parser('recap', description="Generate recap of experiments.",
                                         help="Generate recap of experimets.", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    expe_recap.define_parser(recap_parser)

    bleu_parser = subparsers.add_parser('bleu', description="Compute BLEU score.",
                                        help="Compute BLEU score.", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    bleu_computer.define_parser(bleu_parser)


def do_utils(args):
    func = {"graph": graph_training.do_graph,
            "replace_tgt_unk": replace_tgt_unk.do_replace,
            "recap": expe_recap.do_recap,
            "bleu": bleu_computer.do_bleu
            }[args.__sub_subcommand_name]
    func(args)



#!/usr/bin/env python
"""encoders.py: Implementation of RNNSearch in Chainer"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import chainer
from chainer import cuda, Variable
from chainer import Link, Chain, ChainList
import chainer.functions as F
import chainer.links as L

import rnn_cells

from nmt_chainer.utilities.utils import ortho_init

import logging
logging.basicConfig()
log = logging.getLogger("rnns:encoders")
log.setLevel(logging.INFO)


def make_encoder(Vi, Ei, Hi, init_orth, use_bn_length, cell_type):
    if hasattr(cell_type, "meta_data_cell_type") and cell_type.meta_data_cell_type == rnn_cells.NStepsCell:
        return EncoderNSteps(Vi, Ei, Hi,
                             cell_type=cell_type)
    else:
        return Encoder(Vi, Ei, Hi, init_orth=init_orth, use_bn_length=use_bn_length,
                       cell_type=cell_type)


class EncoderNSteps(Chain):
    """ Chain that encode a sequence.
        The __call_ takes 2 parameters: sequence and mask.
        mask and length should be 2 python lists of same length #length.

        sequence should be a python list of Chainer Variables wrapping a numpy/cupy array of shape (mb_size,) and type int32 each.
            -- where mb_size is the minibatch size
        sequence[i].data[j] should be the jth element of source sequence number i, or a padding value if the sequence number i is
            shorter than j.

        mask should be a python list of Chainer Variables wrapping a numpy/cupy array of shape (mb_size,) and type bool each.
        mask[i].data[j] should be True if and only if sequence[i].data[j] is not a padding value.

        Return a chainer variable of shape (mb_size, #length, 2*Hi) and type float32
    """

    def __init__(self, Vi, Ei, Hi, cell_type):
        gru_f = cell_type(Ei, Hi)
        gru_b = cell_type(Ei, Hi)

        log.info("constructing encoder [%s]" % (cell_type,))
        super(EncoderNSteps, self).__init__(
            emb=L.EmbedID(Vi, Ei),
            #             gru_f = L.GRU(Hi, Ei),
            #             gru_b = L.GRU(Hi, Ei)

            gru_f=gru_f,
            gru_b=gru_b
        )
        self.Hi = Hi

    def __call__(self, sequence, mask, mode="test"):
        assert mode in "test train".split()

        mb_size = sequence[0].data.shape[0]
        max_length_size = len(sequence)

        seq_length = [None] * mb_size
        for num_seq in range(mb_size):
            seq_length[num_seq] = len(sequence) - sum(not mask[i][num_seq] for i in range(len(mask)))

        de_batched_seq = []
        for num_seq in range(mb_size):
            de_batched_seq.append(self.xp.empty((seq_length[num_seq],), dtype=self.xp.int32))
            for i in xrange(seq_length[num_seq]):
                de_batched_seq[-1][i] = sequence[i].data[num_seq]
            de_batched_seq[-1] = Variable(de_batched_seq[-1], volatile="auto")

        embedded_seq = []
        for elem in de_batched_seq:
            embedded_seq.append(self.emb(elem))

        hx, cx, forward_seq = self.gru_f.apply_to_seq(embedded_seq, mode=mode)

        reversed_embedded_seq = []
        for elem in de_batched_seq:
            reversed_embedded_seq.append(self.emb(elem[::-1]))

        hxb, cxb, backward_seq = self.gru_b.apply_to_seq(reversed_embedded_seq, mode=mode)

        assert len(backward_seq) == len(forward_seq) == mb_size

        res = []
        for num_seq in xrange(mb_size):
            assert backward_seq[num_seq].data.shape[0] == forward_seq[num_seq].data.shape[0]
            fb_concatenated = F.concat(
                (forward_seq[num_seq], backward_seq[num_seq][::-1]), 1)
            if forward_seq[num_seq].data.shape[0] < max_length_size:
                pad_length = max_length_size - forward_seq[num_seq].data.shape[0]
                fb_concatenated = F.concat((fb_concatenated,
                                            self.xp.zeros((pad_length, self.Hi * 2), dtype=self.xp.float32)),
                                           axis=0)

            res.append(F.reshape(fb_concatenated, (1, max_length_size, self.Hi * 2)))
        return F.concat(res, axis=0)


class Encoder(Chain):
    """ Chain that encode a sequence.
        The __call_ takes 2 parameters: sequence and mask.
        mask and length should be 2 python lists of same length #length.

        sequence should be a python list of Chainer Variables wrapping a numpy/cupy array of shape (mb_size,) and type int32 each.
            -- where mb_size is the minibatch size
        sequence[i].data[j] should be the jth element of source sequence number i, or a padding value if the sequence number i is
            shorter than j.

        mask should be a python list of Chainer Variables wrapping a numpy/cupy array of shape (mb_size,) and type bool each.
        mask[i].data[j] should be True if and only if sequence[i].data[j] is not a padding value.

        Return a chainer variable of shape (mb_size, #length, 2*Hi) and type float32
    """

    def __init__(self, Vi, Ei, Hi, init_orth=False, use_bn_length=0, cell_type=rnn_cells.LSTMCell):
        gru_f = cell_type(Ei, Hi)
        gru_b = cell_type(Ei, Hi)

        log.info("constructing encoder [%s]" % (cell_type,))
        super(Encoder, self).__init__(
            emb=L.EmbedID(Vi, Ei),
            #             gru_f = L.GRU(Hi, Ei),
            #             gru_b = L.GRU(Hi, Ei)

            gru_f=gru_f,
            gru_b=gru_b
        )
        self.Hi = Hi

        if use_bn_length > 0:
            self.add_link("bn_f", BNList(Hi, use_bn_length))
#             self.add_link("bn_b", BNList(Hi, use_bn_length)) #TODO
        self.use_bn_length = use_bn_length

        if init_orth:
            ortho_init(self.gru_f)
            ortho_init(self.gru_b)

    def __call__(self, sequence, mask, mode="test"):
        assert mode in "test train".split()

        mb_size = sequence[0].data.shape[0]

        mb_initial_states_f = self.gru_f.get_initial_states(mb_size)
        mb_initial_states_b = self.gru_b.get_initial_states(mb_size)

        embedded_seq = []
        for elem in sequence:
            embedded_seq.append(self.emb(elem))

        prev_states = mb_initial_states_f
        forward_seq = []
        for i, x in enumerate(embedded_seq):
            prev_states = self.gru_f(prev_states, x, mode=mode)
            output = prev_states[-1]
            forward_seq.append(output)

        mask_length = len(mask)
        seq_length = len(sequence)
        assert mask_length <= seq_length
        mask_offset = seq_length - mask_length

        prev_states = mb_initial_states_b

        backward_seq = []
        for pos, x in reversed(list(enumerate(embedded_seq))):
            if pos < mask_offset:
                prev_states = self.gru_b(prev_states, x, mode=mode)
                output = prev_states[-1]
            else:
                reshaped_mask = F.broadcast_to(
                    Variable(self.xp.reshape(mask[pos - mask_offset],
                                             (mb_size, 1)), volatile="auto"), (mb_size, self.Hi))

                prev_states = self.gru_b(prev_states, x, mode=mode)
                output = prev_states[-1]

                masked_prev_states = [None] * len(prev_states)
                for num_state in xrange(len(prev_states)):
                    masked_prev_states[num_state] = F.where(reshaped_mask,
                                                            prev_states[num_state], mb_initial_states_b[num_state])  # TODO: optimize?
                prev_states = tuple(masked_prev_states)
                output = prev_states[-1]

            backward_seq.append(output)

        assert len(backward_seq) == len(forward_seq)
        res = []
        for xf, xb in zip(forward_seq, reversed(backward_seq)):
            res.append(F.reshape(F.concat((xf, xb), 1), (-1, 1, 2 * self.Hi)))

        return F.concat(res, 1)



import os
import json
import urllib
import sqlite3
import time
from nmt_chainer.utilities.utils import ensure_path
from nmt_chainer.utilities import bleu_computer
from nmt_chainer.utilities import graph_training
from nmt_chainer.training_module.train_config import load_config_train
from nmt_chainer.translation.eval_config import load_config_eval
from collections import defaultdict
import logging
logging.basicConfig()
log = logging.getLogger("rnns:recap")
log.setLevel(logging.INFO)

data_config_suffix = ".data.config"
train_config_suffix = ".train.config"
eval_config_suffix = ".eval.config.json"


def filenamize(path):
    return urllib.quote_plus(path.replace("/", "xDIRx"))


def process_eval_config(config_fn, dest_dir):
    log.info("processing eval config %s " % config_fn)
    assert config_fn.endswith(eval_config_suffix)

    eval_prefix = config_fn[:-len(eval_config_suffix)]
    log.info("eval prefix: %s " % eval_prefix)

    urlname = filenamize(eval_prefix) + ".html"
    log.info("urlname: %s " % eval_prefix)

    time_config_created = os.path.getmtime(config_fn)

    dest_filename = os.path.join(dest_dir, urlname)

    log.info("create file: %s" % dest_filename)
    f = open(dest_filename, "w")

    config = load_config_eval(config_fn)
#     config = json.load(open(config_fn))
#     ref_fn = config["ref"]

    ref_fn = config.output.ref
    if ref_fn is not None:
        bc_with_unk = bleu_computer.get_bc_from_files(ref_fn, eval_prefix)
        bc_unk_replaced = bleu_computer.get_bc_from_files(
            ref_fn, eval_prefix + ".unk_replaced")
        bleu_unk_replaced = bc_unk_replaced.bleu()
    else:
        bc_with_unk = "#No Ref#"
        bc_unk_replaced = "#No Ref#"
        bleu_unk_replaced = -1

    f.write("<html><body>")
    f.write("Config Filename: %s <p>" % config_fn)
    f.write("eval prefix: %s <p>" % eval_prefix)
    f.write("config file created/modified:%s <p>" % time.ctime(time_config_created))
    f.write("BLEU: %s <p>" % bc_with_unk)
    f.write("BLEU with unk replaced: %s <p>" % bc_unk_replaced)
    for line in open(config_fn):
        f.write(line)
    f.write("</body></html>")

    return urlname, bleu_unk_replaced


def process_data_config(config_fn, dest_dir, data_to_train):
    log.info("processing data config %s " % config_fn)
    assert config_fn.endswith(data_config_suffix)

    data_prefix = config_fn[:-len(data_config_suffix)]
    log.info("data prefix: %s " % data_prefix)

    urlname = filenamize(data_prefix) + ".html"
    log.info("urlname: %s " % data_prefix)

    time_config_created = os.path.getmtime(config_fn)

    dest_filename = os.path.join(dest_dir, urlname)

    log.info("create file: %s" % dest_filename)
    f = open(dest_filename, "w")

    config = json.load(open(config_fn))

    f.write("<html><body>")
    f.write("Config Filename: %s <p>" % config_fn)
    f.write("data prefix: %s <p>" % data_prefix)
    f.write("config file created/modified:%s <p>" % time.ctime(time_config_created))
    for train_urlname in data_to_train[data_prefix]:
        f.write("train: <a href = '../train/%s'>%s</a><p>" % (train_urlname, train_urlname))

    f.write("<pre><code>")
    for line in open(config_fn):
        f.write(line)
    f.write("</code></pre>")
    f.write("</body></html>")

    return urlname, data_prefix, (config["src_fn"], config["tgt_fn"]), time_config_created, config["src_voc_size"], config["tgt_voc_size"]


def process_train_config(config_fn, dest_dir):
    log.info("processing train config %s " % config_fn)
    assert config_fn.endswith(train_config_suffix)

    train_prefix = config_fn[:-len(train_config_suffix)]
    log.info("train prefix: %s " % train_prefix)

    urlname = filenamize(train_prefix) + ".html"
    log.info("urlname: %s " % train_prefix)

    dest_filename = os.path.join(dest_dir, urlname)

    if os.path.exists(dest_filename):
        time_file_modif = os.path.getmtime(dest_filename)
    else:
        time_file_modif = None
    try:
        time_last_exp = os.path.getmtime(train_prefix + ".result.sqlite")
    except OSError:
        time_last_exp = 1

    time_config_created = os.path.getmtime(config_fn)

    if time_file_modif is not None and (time_last_exp + 20) < time_file_modif:
        log.info("skipping creation for %s" % train_prefix)

    log.info("create file: %s" % dest_filename)
    f = open(dest_filename, "w")

    try:
        db = sqlite3.connect(train_prefix + ".result.sqlite")
        c = db.cursor()

        c.execute("SELECT date, bleu_info, iteration, loss, bleu, dev_loss, dev_bleu, avg_time, avg_training_loss FROM exp_data")
        date, bleu_info, iteration, test_loss, test_bleu, dev_loss, dev_bleu, avg_time, avg_training_loss = zip(*list(c.fetchall()))

        infos = dict(
            last_nb_iterations=iteration[-1],
            nb_records=len(test_loss),
            date_last_exp=date[-1],
            best_test_loss=min(test_loss),
            best_dev_loss=min(dev_loss),
            max_test_bleu=max(test_bleu * 100),
            max_dev_bleu=max(dev_bleu * 100),
            best_avg_training_loss=min(avg_training_loss[1:]) if len(avg_training_loss) > 1 else 0,
            avg_time_all=sum(t for t in avg_time if t is not None) / len(avg_time)
        )

    except sqlite3.OperationalError:
        infos = None

    config = load_config_train(config_fn, no_error=True)
#     config = json.load(open(config_fn))

#     data_prefix = config["command_line"]["data_prefix"]
    data_prefix = config.training_management.data_prefix

    try:
        graph_training.generate_graph(sqldb=train_prefix + ".result.sqlite",
                                      #                                   lib = "plotly",
                                      dest=os.path.join(dest_dir, filenamize(train_prefix) + ".graph.html"),
                                      title=train_prefix)
        has_graph = True
    except sqlite3.OperationalError as e:
        log.warn("could not create graph for %s" % train_prefix)
        print e
        has_graph = False

    f.write("<html><body>")
    f.write("Config Filename: %s <p>" % config_fn)
    f.write("train prefix: %s <p>" % train_prefix)

    f.write("data: <a href = '../data/%s'>%s</a><p>" % (filenamize(data_prefix) + ".html", filenamize(data_prefix)))
    if has_graph:
        f.write("<a href = '%s'>graph</a><p>" % (filenamize(train_prefix) + ".graph.html"))

    f.write("config file created/modified:%s <p>" % time.ctime(time_config_created))
    f.write("last exp:%s <p>" % time.ctime(time_last_exp))
    time_file_modif_new = os.path.getmtime(dest_filename)
    f.write("this file created/modified:%s <p>" % time.ctime(time_file_modif_new))

    if infos is not None:
        for key in sorted(infos.keys()):
            f.write("%s : %r <p>" % (key, infos[key]))

    f.write("<pre><code>")
    for line in open(config_fn):
        f.write(line)
    f.write("</code></pre>")
    f.write("</body></html>")

    return urlname, data_prefix, time_last_exp, infos


def define_parser(parser):
    parser.add_argument("source_dir", help="directory containing experiments results (also look into subdirectories)")
    parser.add_argument("target_dir", help="directory where html files will be generated")


def do_recap(args):
    data_dir = os.path.join(args.target_dir, "data")
    train_dir = os.path.join(args.target_dir, "train")
    eval_dir = os.path.join(args.target_dir, "eval")

    ensure_path(args.target_dir)
    ensure_path(data_dir)
    ensure_path(train_dir)
    ensure_path(eval_dir)

    index = open(os.path.join(args.target_dir, "index.html"), "w")
    index.write("<html><body>")

    data_urlname_list = defaultdict(list)
    train_urlname_list = defaultdict(list)

    data_config_fn_list = []
    eval_config_fn_list = []

    itdir = os.walk(args.source_dir)

    data_to_train = defaultdict(list)
    train_to_data = {}
    for current_dir, dirs, files in itdir:
        for fn in files:
            if fn.endswith(train_config_suffix):
                fn_full = os.path.join(current_dir, fn)
                urlname, data_prefix, time_last_exp, infos = process_train_config(
                    fn_full, train_dir)
                data_to_train[data_prefix].append(urlname)
                train_to_data[urlname] = data_prefix
                train_urlname_list[data_prefix].append(
                    (time_last_exp, urlname, infos))
            elif fn.endswith(data_config_suffix):
                fn_full = os.path.join(current_dir, fn)
                data_config_fn_list.append(fn_full)
            elif fn.endswith(eval_config_suffix):
                fn_full = os.path.join(current_dir, fn)
                eval_config_fn_list.append(fn_full)
            else:
                pass

    data_to_srctgt = {}
    for fn_full in data_config_fn_list:
        urlname, data_prefix, src_tgt_fn, time_config_created, src_voc_size, tgt_voc_size = process_data_config(fn_full, data_dir, data_to_train)
        data_urlname_list[src_tgt_fn].append((time_config_created, urlname, data_prefix, src_voc_size, tgt_voc_size))
        data_to_srctgt[data_prefix] = src_tgt_fn

    index.write("<h1>DATA</h1><p>")
    for src_tgt_fn, urlname_list in data_urlname_list.iteritems():
        index.write("<h3>** src: %s | tgt: %s **</h3>" % src_tgt_fn)
        urlname_list.sort(reverse=True)
        for time_config_created, urlname, data_prefix, src_voc_size, tgt_voc_size in urlname_list:
            index.write('%s s:%i t:%i \t<a href = "data/%s">%s</a><p/>' % (time.ctime(time_config_created),
                                                                           src_voc_size, tgt_voc_size, urlname, data_prefix))
    train_urlname_list_src_tgt = defaultdict(list)
    for data_path, urlname_list in train_urlname_list.iteritems():
        if data_path in data_to_srctgt:
            train_urlname_list_src_tgt[data_to_srctgt[data_path]
                                       ] += urlname_list
        else:
            train_urlname_list_src_tgt[("unk", "unk")] += urlname_list

    current_time = time.time()
    index.write("<h1>TRAIN</h1><p>")
    for src_tgt_fn in sorted(train_urlname_list_src_tgt.keys(), key=lambda x: max(train_urlname_list_src_tgt[x][i][0] for i in xrange(len(train_urlname_list_src_tgt[x]))), reverse=True):
        urlname_list = train_urlname_list_src_tgt[src_tgt_fn]
        index.write("<h3>** src: %s | tgt: %s **</h3>" % src_tgt_fn)
        urlname_list.sort(reverse=True)
        for time_last_exp, urlname, infos in urlname_list:
            if abs(time_last_exp - current_time) < 3000:
                recently_updated = True
            else:
                recently_updated = False
            if recently_updated:
                timestring = "<b>%s [RCT]</b>" % time.ctime(time_last_exp)
            else:
                timestring = "%s" % time.ctime(time_last_exp)
            index.write('%s <a href = "train/%s">%s</a><p/>' % (timestring, urlname, urlname.split("xDIRx")[-1]))
            if infos is not None:
                for key in sorted(infos.keys()):
                    index.write("%s : %r  ||| " % (key, infos[key]))
            index.write("<p>")

    index.write("<h1>EVAL</h1><p>")
    for fn_full in eval_config_fn_list:
        urlname, bleu = process_eval_config(fn_full, eval_dir)
        index.write('<a href = "eval/%s">%s</a> %f <p/>' % (urlname, urlname.split("xDIRx")[-1], bleu))

    index.write("</body></html>")


def commandline():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    define_parser(parser)
    args = parser.parse_args()
    do_recap(args)


if __name__ == '__main__':
    commandline()



import logging

logging.basicConfig()
log = logging.getLogger("rnns:indexer")
log.setLevel(logging.INFO)


class Indexer(object):

    def __init__(self, unk_tag="#UNK#"):
        self.dic = {}
        self.lst = []
        self.unk_label_dictionary = None
        self.finalized = False

    def add_word(self, w, should_be_new=False, should_not_be_int=True):
        assert not self.finalized
        assert not (should_not_be_int and isinstance(w, int))
        if w not in self.dic:
            new_idx = len(self.lst)
            self.dic[w] = new_idx
            self.lst.append(w)
            assert len(self.lst) == len(self.dic)
        else:
            assert not should_be_new

#     def assign_index_to_voc(self, voc_iter, all_should_be_new=False):
#         assert not self.finalized
#         for w in voc_iter:
#             self.add_word(w, should_be_new=all_should_be_new)

    def finalize(self):
        assert not self.finalized
        assert len(self.dic) == len(self.lst)
        self.add_word(0, should_be_new=False, should_not_be_int=False)
        self.finalized = True

    def get_one_unk_idx(self, w):
        assert self.finalized
        assert w not in self.dic
        if self.unk_label_dictionary is not None:
            return self.dic[self.unk_label_dictionary.get(w, 1)]
        else:
            return self.dic[0]

    def is_unk_idx(self, idx):
        assert self.finalized
        assert idx < len(self.lst)
        return isinstance(self.lst[idx], int)

#     def add_unk_label_dictionary(self, unk_dic):
#         assert not self.finalized
#         self.unk_label_dictionary = unk_dic

    def convert(self, seq):
        assert self.finalized
        assert len(self.dic) == len(self.lst)
        res = [None] * len(seq)
        for pos, w in enumerate(seq):
            assert not isinstance(w, int)
            if w in self.dic:
                idx = self.dic[w]
            else:
                idx = self.get_one_unk_idx(w)
            res[pos] = idx
        return res

#     def convert_and_update_unk_tags(self, seq, give_unk_label):
#         assert not self.finalized
#         assert len(self.dic) == len(self.lst)
#         res = [None] * len(seq)
#         for pos, w in enumerate(seq):
#             assert not isinstance(w, int)
#             if w in self.dic:
#                 idx = self.dic[w]
#             else:
#                 aligned_pos = give_unk_label(pos, w)
#                 if aligned_pos not in self.dic:
#                     self.add_word(aligned_pos, should_be_new=True,
#                                   should_not_be_int=False)
#                 idx = self.dic[aligned_pos]
#             res[pos] = idx
#         return res

    def deconvert(self, seq, unk_tag="#UNK#", no_oov=True, eos_idx=None):
        assert self.finalized
        assert eos_idx is None or eos_idx >= len(self.lst)
        res = []
        for num, idx in enumerate(seq):
            if idx >= len(self.lst):
                if eos_idx is not None and eos_idx == idx:
                    w = "#EOS#"
                elif no_oov:
                    raise KeyError()
                else:
                    log.warn("unknown idx: %i / %i" % (idx, len(self.lst)))
                    continue
            else:
                w = self.lst[idx]

            if isinstance(w, int):
                if callable(unk_tag):
                    w = unk_tag(num, w)
                else:
                    w = unk_tag

            res.append(w)
        return res

    def __len__(self):
        assert self.finalized
        assert len(self.dic) == len(self.lst)
        return len(self.lst)

    def to_serializable(self):
        return {"type": "simple_indexer", "rev": 1, "voc_lst": self.lst, "unk_label_dic": self.unk_label_dictionary}

    @staticmethod
    def make_from_serializable(datas):
        if isinstance(datas, list):
            # legacy mode
            log.info("loading legacy voc")
            voc_lst = datas
            assert 0 not in voc_lst
            res = Indexer()
            res.lst = list(voc_lst)  # add UNK
            for idx, w in enumerate(res.lst):
                assert isinstance(w, basestring)
                res.dic[w] = idx
            assert len(res.dic) == len(res.lst)
            res.finalize()
            return res
        else:
            assert isinstance(datas, dict)
            assert datas["type"] == "simple_indexer"
            assert datas["rev"] == 1
            voc_lst = datas["voc_lst"]
            res = Indexer()
            res.lst = list(voc_lst)
            res.unk_label_dictionary = datas["unk_label_dic"]
            for idx, w in enumerate(voc_lst):
                res.dic[w] = idx
            res.finalized = True
            return res

    @staticmethod
    def check_if_data_indexer(datas):
        if isinstance(datas, list):
            return True
        elif isinstance(datas, dict) and "type" in datas and datas["type"] == "simple_indexer":
            return True

# class Indexer(object):
#     def __init__(self, unk_tag = "#UNK#"):
#         self.dic = {}
#         self.lst = []
#         self.finalized = False
#
#     def add_word(self, w, should_be_new = False):
#         assert not self.finalized
#         assert w is not None
#         if w not in self.dic:
#             new_idx = len(self.lst)
#             self.dic[w] = new_idx
#             self.lst.append(w)
#             assert len(self.lst) == len(self.dic)
#         else:
#             assert not should_be_new
#
#     def assign_index_to_voc(self, voc_iter, all_should_be_new = False):
#         assert not self.finalized
#         for w in voc_iter:
#             self.add_word(w, should_be_new = all_should_be_new)
#
#     def finalize(self):
#         assert not self.finalized
#         self.dic[None] = len(self.lst)
#         self.lst.append(None)
#         self.finalized = True
#
#     def get_unk_idx(self):
#         assert self.finalized
#         return self.dic[None]
#
#     def convert(self, seq):
#         assert self.finalized
#         assert len(self.dic) == len(self.lst)
#         unk_idx = self.get_unk_idx()
# #         res = np.empty( (len(seq),), dtype = np.int32)
#         res = [None] * len(seq)
#         for pos, w in enumerate(seq):
#             assert w is not None
#             idx = self.dic.get(w, unk_idx)
#             res[pos] = idx
#         return res
#
#     def deconvert(self, seq, unk_tag = "#UNK#", no_oov = True, eos_idx = None):
#         assert self.finalized
#         assert eos_idx is None or eos_idx >= len(self.lst)
#         res = []
#         for num, idx in enumerate(seq):
#             if idx >= len(self.lst):
#                 if eos_idx is not None and eos_idx == idx:
#                     w = "#EOS#"
#                 elif no_oov:
#                     raise KeyError()
#                 else:
#                     log.warn("unknown idx: %i / %i"%(idx, len(self.lst)))
#                     continue
#             else:
#                 w = self.lst[idx]
#             if w is None:
#                 if callable(unk_tag):
#                     w = unk_tag(num)
#                 else:
#                     w = unk_tag
#             res.append(w)
#         return res
#
#     def __len__(self):
#         assert self.finalized
#         assert len(self.dic) == len(self.lst)
#         return len(self.lst)
#
#     def to_serializable(self):
#         return {"type": "simple_indexer", "rev": 1,  "voc_lst" : self.lst}
#
#     @staticmethod
#     def make_from_serializable(datas):
#         assert datas["type"] == "simple_indexer"
#         assert datas["rev"] == 1
#         voc_lst = datas["voc_lst"]
#         res = Indexer()
#         res.lst = list(voc_lst)
#         for idx, w in enumerate(voc_lst):
#             res.dic[w] = idx
#         res.finalized = True
#         return res



import numpy
import six

from chainer import cuda
from chainer import function
from chainer.utils import array
from chainer.utils import type_check

from chainer.functions.math.matmul import _get_batch_mat_shape, _check_ndim, _convert_type, _get_check_index, _matmul

try:
    from chainer.functions.math.matmul import _batch_matmul_gpu
except ImportError:
    def _batch_matmul_gpu(*args, **kwds):
        # TO DO: fix this for recent chainer
        print "Sorry, use of lexical probabilities is broken with this version of chainer"
        raise NotImplemented


class MatMulConstant(function.Function):

    def __init__(self, b, transa=False, transb=False):
        self.transa = transa
        self.transb = transb
        self.b = b

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 1)
        a_type, = in_types

        type_check.expect(
            a_type.dtype.kind == 'f',
        )

        _check_ndim(a_type)

        a_type = _convert_type(a_type)
        a_idx = _get_check_index(self.transa, False)

    def forward(self, x):
        a, = x
        return _matmul(a, self.b, transa=self.transa, transb=self.transb),

    def backward(self, x, gy):
        gx0 = _matmul(
            gy[0], self.b, transb=not self.transb, transout=self.transa
        ).reshape(x[0].shape)
        return gx0,


def matmul_constant(a, b, transa=False, transb=False):
    """Computes the matrix multiplication of two arrays.
    Args:
        a (Variable): The left operand of the matrix multiplication.
            A 1-D array of shape ``(N,)`` is considered as an
            :math:`N \\times 1` matrix.
            A 2-D array of shape ``(M, N)`` is considered as an
            :math:`M \\times N` matrix.
        b (Variable): The right operand of the matrix multiplication.
            Its array is treated as a matrix in the same way as ``a``'s array.
        transa (bool): If ``True``, transpose a.
        transb (bool): If ``True``, transpose b.
    Returns:
        ~chainer.Variable: The result of the matrix multiplication as a 2-D
            array.
    """
    return MatMulConstant(b, transa=transa, transb=transb)(a)


class BatchMatMulConstant(function.Function):
    def __init__(self, b, transa=False, transb=False):
        self.transa = transa
        self.transb = transb
        self.b = b

    def _output_shape(self, a, b):
        batch_size = len(a)
        m = _get_batch_mat_shape(a.shape)[2 if self.transa else 1]
        n = _get_batch_mat_shape(b.shape)[1 if self.transb else 2]
        return batch_size, m, n

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 1)
        a_type, = in_types

        type_check.expect(
            a_type.dtype == numpy.float32
            #             ,
            #             self.b.dtype == numpy.float32
        )

        _check_ndim(a_type, lower=2, upper=3)
#         _check_ndim(b_type, lower=2, upper=3)

        a_type = _convert_type(a_type, vector_ndim=2)
#         b_type = _convert_type(b_type, vector_ndim=2)
        a_idx = _get_check_index(self.transa, False, row_idx=1, col_idx=2)
#         b_idx = _get_check_index(self.transb, True, row_idx=1, col_idx=2)
#         type_check.expect(
#             a_type.shape[a_idx] == b_type.shape[b_idx]
#         )

    def forward_cpu(self, x):
        a, = x
        b = self.b
        batch_size = a.shape[0]
        shape = self._output_shape(a, b)
        ret_dtype = numpy.find_common_type([a.dtype, b.dtype], [])
        ret = numpy.empty(shape, dtype=ret_dtype)
        for i in six.moves.range(batch_size):
            ret[i] = _matmul(
                a[i], b[i], transa=self.transa, transb=self.transb)
        return ret,

    def backward_cpu(self, x, gy):
        a, = x
        b = self.b
        ga = numpy.empty_like(a)
        a0shape = a[0].shape
        for i in six.moves.range(len(a)):
            ga[i] = _matmul(
                gy[0][i], b[i], transb=not self.transb,
                transout=self.transa).reshape(a0shape)
        return ga,

    def forward_gpu(self, x):
        a, = x
        b = self.b
        shape = self._output_shape(a, b)
        ret = cuda.cupy.zeros(shape, dtype=a.dtype)
        _batch_matmul_gpu(
            a, b, transa=self.transa, transb=self.transb, out=ret)
        return ret,

    def backward_gpu(self, x, gy):
        a, = x
        b = self.b
        ga = cuda.cupy.empty(_get_batch_mat_shape(a.shape), a.dtype)
        _batch_matmul_gpu(
            gy[0], b, transb=not self.transb, transout=self.transa, out=ga)
        ga = ga.reshape(a.shape)
        return ga,


def batch_matmul_constant(a, b, transa=False, transb=False):
    """Computes the batch matrix multiplications of two sets of arrays.

    Args:
        a (Variable): The left operand of the batch matrix multiplications.
            A 2-D array of shape ``(B, N)`` is considered as B
            :math:`N \\times 1` matrices.
            A 3-D array of shape ``(B, M, N)`` is considered as B
            :math:`M \\times N` matrices.
        b (Variable): The right operand of the batch matrix multiplications.
            Its array is treated as matrices in the same way as ``a``'s array.
        transa (bool): If ``True``, transpose each matrix in a.
        transb (bool): If ``True``, transpose each matrix in b.

    Returns:
        ~chainer.Variable: The result of the batch matrix multiplications as a
            3-D array.
    """
    return BatchMatMulConstant(b, transa=transa, transb=transb)(a)



#!/usr/bin/env python
"""eval.py: Use a RNNSearch Model"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import json
import numpy as np
from chainer import cuda, serializers
import sys
from nmt_chainer.dataprocessing.processors import build_dataset_one_side_pp
import nmt_chainer.dataprocessing.make_data as make_data
import nmt_chainer.training_module.train as train
import nmt_chainer.training_module.train_config as train_config
import re

# from utils import make_batch_src, make_batch_src_tgt, minibatch_provider, compute_bleu_with_unk_as_wrong, de_batch
from nmt_chainer.translation.evaluation import (greedy_batch_translate,
                                                #                         convert_idx_to_string,
                                                batch_align,
                                                beam_search_translate,
                                                #                         convert_idx_to_string_with_attn
                                                )

# import visualisation
from nmt_chainer.utilities import bleu_computer
import logging
import codecs
# import h5py
import bokeh.embed


logging.basicConfig()
log = logging.getLogger("rnns:eval")
log.setLevel(logging.INFO)


class AttentionVisualizer(object):
    def __init__(self):
        self.plots_list = []

    def add_plot(self, src_w, tgt_w, attn, include_sum=True, visual_attribs=None):
        from nmt_chainer.utilities import visualisation
        alignment = np.zeros((len(src_w) + 1, len(tgt_w)))
        sum_al = [0] * len(tgt_w)
        for i in xrange(len(src_w)):
            for j in xrange(len(tgt_w)):
                alignment[i, j] = attn[j][i]
                sum_al[j] += alignment[i, j]
        for j in xrange(len(tgt_w)):
            alignment[len(src_w), j] = sum_al[j]

        src = src_w
        if include_sum:
            src += ["SUM_ATTN"]
        if visual_attribs is not None:
            p1 = visualisation.make_alignment_figure(
                src, tgt_w, alignment, **visual_attribs)
        else:
            p1 = visualisation.make_alignment_figure(src, tgt_w, alignment)

        self.plots_list.append(p1)

    def make_plot(self, output_file):
        from nmt_chainer.utilities import visualisation
        log.info("writing attention to {0}".format(output_file))
        p_all = visualisation.Column(*self.plots_list)
        if isinstance(output_file, tuple):
            script_output_fn, div_output_fn = output_file
            script, div = bokeh.embed.components(p_all)
            with open(script_output_fn, 'w') as f:
                f.write(script.encode('utf-8'))
            with open(div_output_fn, 'w') as f:
                f.write(div)
        else:
            visualisation.output_file(output_file)
            visualisation.show(p_all)


class RichOutputWriter(object):
    def __init__(self, filename):
        log.info("writing JSON translation infos to %s" % filename)
        self.filename = filename
        self.output = open(filename, "w")
        self.no_entry_yet = True
        self.output.write("[\n")

    def add_info(self, src, translated, t, score, attn, unk_mapping=None):
        if not self.no_entry_yet:
            self.output.write(",\n")
        else:
            self.no_entry_yet = False
        self.output.write("{\"tr\": ")
        self.output.write(json.dumps(translated))
        self.output.write(",\n\"attn\": ")
        self.output.write(json.dumps(
            [[float(a) for a in a_list] for a_list in attn]))
        if unk_mapping is not None:
            self.output.write(",\n\"unk_mapping\": ")
            self.output.write(json.dumps(unk_mapping))
        self.output.write("}")

    def finish(self):
        self.output.write("\n]")
        self.output.close()
        log.info("done writing JSON translation infos to %s" % self.filename)


def beam_search_all(gpu, encdec, eos_idx, src_data, beam_width, beam_pruning_margin, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, nb_steps,
                    nb_steps_ratio, beam_score_length_normalization, beam_score_length_normalization_strength, post_score_length_normalization, post_score_length_normalization_strength,
                    post_score_coverage_penalty, post_score_coverage_penalty_strength,
                    groundhog,
                    tgt_unk_id, tgt_indexer, force_finish=False,
                    prob_space_combination=False, reverse_encdec=None,
                    use_unfinished_translation_if_none_found=False,
                    replace_unk=False,
                    src=None,
                    dic=None,
                    remove_unk=False,
                    normalize_unicode_unk=False,
                    attempt_to_relocate_unk_source=False):

    log.info("starting beam search translation of %i sentences" % len(src_data))
    if isinstance(encdec, (list, tuple)) and len(encdec) > 1:
        log.info("using ensemble of %i models" % len(encdec))

    with cuda.get_device(gpu):
        translations_gen = beam_search_translate(
            encdec, eos_idx, src_data, beam_width=beam_width, nb_steps=nb_steps,
            gpu=gpu, beam_pruning_margin=beam_pruning_margin,
            beam_score_coverage_penalty=beam_score_coverage_penalty,
            beam_score_coverage_penalty_strength=beam_score_coverage_penalty_strength,
            nb_steps_ratio=nb_steps_ratio,
            need_attention=True,
            beam_score_length_normalization=beam_score_length_normalization,
            beam_score_length_normalization_strength=beam_score_length_normalization_strength,
            post_score_length_normalization=post_score_length_normalization,
            post_score_length_normalization_strength=post_score_length_normalization_strength,
            post_score_coverage_penalty=post_score_coverage_penalty,
            post_score_coverage_penalty_strength=post_score_coverage_penalty_strength,
            groundhog=groundhog, force_finish=force_finish,
            prob_space_combination=prob_space_combination,
            reverse_encdec=reverse_encdec,
            use_unfinished_translation_if_none_found=use_unfinished_translation_if_none_found)

        for num_t, (t, score, attn) in enumerate(translations_gen):
            if num_t % 200 == 0:
                print >>sys.stderr, num_t,
            elif num_t % 40 == 0:
                print >>sys.stderr, "*",
#                 t, score = bests[1]
#                 ct = convert_idx_to_string(t, tgt_voc + ["#T_UNK#"])
#                 ct = convert_idx_to_string_with_attn(t, tgt_voc, attn, unk_idx = len(tgt_voc))
            if tgt_unk_id == "align":
                def unk_replacer(num_pos, unk_id):
                    unk_pattern = "#T_UNK_%i#"
                    a = attn[num_pos]
                    xp = cuda.get_array_module(a)
                    src_pos = int(xp.argmax(a))
                    return unk_pattern % src_pos
            elif tgt_unk_id == "id":
                def unk_replacer(num_pos, unk_id):
                    unk_pattern = "#T_UNK_%i#"
                    return unk_pattern % unk_id
            else:
                assert False

            translated = tgt_indexer.deconvert_swallow(t, unk_tag=unk_replacer)

            unk_mapping = []
            ct = " ".join(translated)
            if ct != '':
                unk_pattern = re.compile("#T_UNK_(\d+)#")
                for idx, word in enumerate(ct.split(' ')):
                    match = unk_pattern.match(word)
                    if (match):
                        unk_mapping.append(match.group(1) + '-' + str(idx))

                if replace_unk:
                    from nmt_chainer.utilities import replace_tgt_unk
                    translated = replace_tgt_unk.replace_unk_from_string(ct, src, dic, remove_unk, normalize_unicode_unk, attempt_to_relocate_unk_source).strip().split(" ")

            yield src_data[num_t], translated, t, score, attn, unk_mapping
        print >>sys.stderr


def translate_to_file_with_beam_search(dest_fn, gpu, encdec, eos_idx, src_data, beam_width, beam_pruning_margin, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, nb_steps,
                                       nb_steps_ratio, beam_score_length_normalization, beam_score_length_normalization_strength, post_score_length_normalization, post_score_length_normalization_strength,
                                       post_score_coverage_penalty, post_score_coverage_penalty_strength,
                                       groundhog,
                                       tgt_unk_id, tgt_indexer, force_finish=False,
                                       prob_space_combination=False, reverse_encdec=None,
                                       generate_attention_html=None, attn_graph_with_sum=True, attn_graph_attribs=None, src_indexer=None, rich_output_filename=None,
                                       use_unfinished_translation_if_none_found=False,
                                       replace_unk=False,
                                       src=None,
                                       dic=None,
                                       remove_unk=False,
                                       normalize_unicode_unk=False,
                                       attempt_to_relocate_unk_source=False):

    log.info("writing translation to %s " % dest_fn)
    out = codecs.open(dest_fn, "w", encoding="utf8")

    translation_iterator = beam_search_all(gpu, encdec, eos_idx, src_data, beam_width, beam_pruning_margin, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, nb_steps,
                                           nb_steps_ratio, beam_score_length_normalization, beam_score_length_normalization_strength, post_score_length_normalization, post_score_length_normalization_strength,
                                           post_score_coverage_penalty, post_score_coverage_penalty_strength,
                                           groundhog,
                                           tgt_unk_id, tgt_indexer, force_finish=force_finish,
                                           prob_space_combination=prob_space_combination, reverse_encdec=reverse_encdec,
                                           use_unfinished_translation_if_none_found=use_unfinished_translation_if_none_found,
                                           replace_unk=replace_unk,
                                           src=src,
                                           dic=dic,
                                           remove_unk=remove_unk,
                                           normalize_unicode_unk=normalize_unicode_unk,
                                           attempt_to_relocate_unk_source=attempt_to_relocate_unk_source)

    attn_vis = None
    if generate_attention_html is not None:
        attn_vis = AttentionVisualizer()
        assert src_indexer is not None

    rich_output = None
    if rich_output_filename is not None:
        rich_output = RichOutputWriter(rich_output_filename)

    for src, translated, t, score, attn, unk_mapping in translation_iterator:
        if rich_output is not None:
            rich_output.add_info(src, translated, t, score, attn, unk_mapping=unk_mapping)
        if attn_vis is not None:
            attn_vis.add_plot(
                src_indexer.deconvert_swallow(src),
                translated,
                attn,
                attn_graph_with_sum,
                attn_graph_attribs)
        ct = tgt_indexer.deconvert_post(translated)
        out.write(ct + "\n")

    if rich_output is not None:
        rich_output.finish()

    if attn_vis is not None:
        attn_vis.make_plot(generate_attention_html)


def create_and_load_encdec_from_files(config_training_fn, trained_model):
    log.info("loading model config from %s" % config_training_fn)

    config_training = train_config.load_config_train(config_training_fn)
    encdec, eos_idx, src_indexer, tgt_indexer = train.create_encdec_and_indexers_from_config_dict(config_training)

    log.info("loading model from %s" % trained_model)
    serializers.load_npz(trained_model, encdec)

    return encdec, eos_idx, src_indexer, tgt_indexer


def check_if_vocabulary_info_compatible(this_eos_idx, this_src_indexer, this_tgt_indexer, eos_idx, src_indexer, tgt_indexer):
    if eos_idx != this_eos_idx:
        raise Exception("incompatible models")
    if len(src_indexer) != len(this_src_indexer):
        raise Exception("incompatible models")
    if len(tgt_indexer) != len(this_tgt_indexer):
        raise Exception("incompatible models")


def create_encdec(config_eval):
    encdec_list = []
    eos_idx, src_indexer, tgt_indexer = None, None, None

    if config_eval.training_config is not None:
        assert config_eval.trained_model is not None
        encdec, eos_idx, src_indexer, tgt_indexer = create_and_load_encdec_from_files(
            config_eval.training_config, config_eval.trained_model)

        encdec_list.append(encdec)

    if 'load_model_config' in config_eval.process and config_eval.process.load_model_config is not None:
        for config_filename in config_eval.process.load_model_config:
            log.info(
                "loading model and parameters from config %s" %
                config_filename)
            config_training = train_config.load_config_train(config_filename)
            encdec, this_eos_idx, this_src_indexer, this_tgt_indexer = train.create_encdec_and_indexers_from_config_dict(config_training, load_config_model="yes")

            if eos_idx is None:
                assert len(encdec_list) == 0
                assert src_indexer is None
                assert tgt_indexer is None
                eos_idx, src_indexer, tgt_indexer = this_eos_idx, this_src_indexer, this_tgt_indexer
            else:
                check_if_vocabulary_info_compatible(this_eos_idx, this_src_indexer, this_tgt_indexer, eos_idx, src_indexer, tgt_indexer)

            encdec_list.append(encdec)

    assert len(encdec_list) > 0

    if 'additional_training_config' in config_eval.process and config_eval.process.additional_training_config is not None:
        assert len(config_eval.process.additional_training_config) == len(config_eval.process.additional_trained_model)

        for (config_training_fn, trained_model_fn) in zip(config_eval.process.additional_training_config,
                                                          config_eval.process.additional_trained_model):
            this_encdec, this_eos_idx, this_src_indexer, this_tgt_indexer = create_and_load_encdec_from_files(
                config_training_fn, trained_model_fn)

            check_if_vocabulary_info_compatible(this_eos_idx, this_src_indexer, this_tgt_indexer, eos_idx, src_indexer, tgt_indexer)

#             if args.gpu is not None:
#                 this_encdec = this_encdec.to_gpu(args.gpu)

            encdec_list.append(this_encdec)

    if 'gpu' in config_eval.process and config_eval.process.gpu is not None:
        encdec_list = [encdec.to_gpu(config_eval.process.gpu) for encdec in encdec_list]

    if 'reverse_training_config' in config_eval.process and config_eval.process.reverse_training_config is not None:
        reverse_encdec, reverse_eos_idx, reverse_src_indexer, reverse_tgt_indexer = create_and_load_encdec_from_files(
            config_eval.process.reverse_training_config, config_eval.process.reverse_trained_model)

        if eos_idx != reverse_eos_idx:
            raise Exception("incompatible models")

        if len(src_indexer) != len(reverse_src_indexer):
            raise Exception("incompatible models")

        if len(tgt_indexer) != len(reverse_tgt_indexer):
            raise Exception("incompatible models")

        if config_eval.process.gpu is not None:
            reverse_encdec = reverse_encdec.to_gpu(config_eval.process.gpu)
    else:
        reverse_encdec = None

    return encdec_list, eos_idx, src_indexer, tgt_indexer, reverse_encdec


def do_eval(config_eval):
    src_fn = config_eval.process.src_fn
    tgt_fn = config_eval.output.tgt_fn
    mode = config_eval.method.mode
    gpu = config_eval.process.gpu
    dest_fn = config_eval.process.dest_fn
    mb_size = config_eval.process.mb_size
    nb_steps = config_eval.method.nb_steps
    nb_steps_ratio = config_eval.method.nb_steps_ratio
    max_nb_ex = config_eval.process.max_nb_ex
    nbest_to_rescore = config_eval.output.nbest_to_rescore

    beam_width = config_eval.method.beam_width
    beam_pruning_margin = config_eval.method.beam_pruning_margin
    beam_score_length_normalization = config_eval.method.beam_score_length_normalization
    beam_score_length_normalization_strength = config_eval.method.beam_score_length_normalization_strength
    beam_score_coverage_penalty = config_eval.beam_score_coverage_penalty
    beam_score_coverage_penalty_strength = config_eval.beam_score_coverage_penalty_strength
    post_score_length_normalization = config_eval.method.post_score_length_normalization
    post_score_length_normalization_strength = config_eval.method.post_score_length_normalization_strength
    groundhog = config_eval.method.groundhog
    tgt_unk_id = config_eval.output.tgt_unk_id
    force_finish = config_eval.method.force_finish
    prob_space_combination = config_eval.method.prob_space_combination
    generate_attention_html = config_eval.output.generate_attention_html
    rich_output_filename = config_eval.output.rich_output_filename

    ref = config_eval.output.ref
    dic = config_eval.output.dic
    normalize_unicode_unk = config_eval.output.normalize_unicode_unk
    attempt_to_relocate_unk_source = config_eval.output.attempt_to_relocate_unk_source
    remove_unk = config_eval.output.remove_unk

    post_score_coverage_penalty = config_eval.method.post_score_coverage_penalty
    post_score_coverage_penalty_strength = config_eval.method.post_score_coverage_penalty_strength

    if dest_fn is not None:
        save_eval_config_fn = dest_fn + ".eval.config.json"
        log.info("Saving eval config to %s" % save_eval_config_fn)
        config_eval.save_to(save_eval_config_fn)
#     json.dump(config_eval, open(save_eval_config_fn, "w"), indent=2, separators=(',', ': '))

    encdec_list, eos_idx, src_indexer, tgt_indexer, reverse_encdec = create_encdec(config_eval)

    if src_fn is not None:
        log.info("opening source file %s" % src_fn)
        src_data, stats_src_pp = build_dataset_one_side_pp(src_fn, src_pp=src_indexer,
                                                           max_nb_ex=max_nb_ex)
        log.info("src data stats:\n%s", stats_src_pp.make_report())

#     log.info("%i sentences loaded" % make_data_infos.nb_ex)
#     log.info("#tokens src: %i   of which %i (%f%%) are unknown"%(make_data_infos.total_token,
#                                                                  make_data_infos.total_count_unk,
#                                                                  float(make_data_infos.total_count_unk * 100) /
#                                                                     make_data_infos.total_token))

    tgt_data = None
    if tgt_fn is not None:
        log.info("opening target file %s" % tgt_fn)
        tgt_data, stats_tgt_pp = build_dataset_one_side_pp(tgt_fn, src_pp=tgt_indexer,
                                                           max_nb_ex=max_nb_ex)
        log.info("tgt data stats:\n%s", stats_tgt_pp.make_report())
#         log.info("%i sentences loaded"%make_data_infos.nb_ex)
#         log.info("#tokens src: %i   of which %i (%f%%) are unknown"%(make_data_infos.total_token,
#                                                                  make_data_infos.total_count_unk,
#                                                                  float(make_data_infos.total_count_unk * 100) /
#                                                                     make_data_infos.total_token))


#     translations = greedy_batch_translate(encdec, eos_idx, src_data, batch_size = mb_size, gpu = args.gpu)

    if mode == "translate":
        log.info("writing translation of to %s" % dest_fn)
        with cuda.get_device(gpu):
            assert len(encdec_list) == 1
            translations = greedy_batch_translate(
                encdec_list[0], eos_idx, src_data, batch_size=mb_size, gpu=gpu, nb_steps=nb_steps)
        out = codecs.open(dest_fn, "w", encoding="utf8")
        for t in translations:
            if t[-1] == eos_idx:
                t = t[:-1]
            ct = tgt_indexer.deconvert(t, unk_tag="#T_UNK#")
#             ct = convert_idx_to_string(t, tgt_voc + ["#T_UNK#"])
            out.write(ct + "\n")

    elif mode == "beam_search":
        if config_eval.process.server is not None:
            from nmt_chainer.translation.server import do_start_server
            do_start_server(config_eval)
        else:
            translate_to_file_with_beam_search(dest_fn, gpu, encdec_list, eos_idx, src_data,
                                               beam_width=beam_width,
                                               beam_pruning_margin=beam_pruning_margin,
                                               beam_score_coverage_penalty=beam_score_coverage_penalty,
                                               beam_score_coverage_penalty_strength=beam_score_coverage_penalty_strength,
                                               nb_steps=nb_steps,
                                               nb_steps_ratio=nb_steps_ratio,
                                               beam_score_length_normalization=beam_score_length_normalization,
                                               beam_score_length_normalization_strength=beam_score_length_normalization_strength,
                                               post_score_length_normalization=post_score_length_normalization,
                                               post_score_length_normalization_strength=post_score_length_normalization_strength,
                                               post_score_coverage_penalty=post_score_coverage_penalty,
                                               post_score_coverage_penalty_strength=post_score_coverage_penalty_strength,
                                               groundhog=groundhog,
                                               tgt_unk_id=tgt_unk_id,
                                               tgt_indexer=tgt_indexer,
                                               force_finish=force_finish,
                                               prob_space_combination=prob_space_combination,
                                               reverse_encdec=reverse_encdec,
                                               generate_attention_html=generate_attention_html,
                                               src_indexer=src_indexer,
                                               rich_output_filename=rich_output_filename,
                                               use_unfinished_translation_if_none_found=True)

    elif mode == "eval_bleu":
        #         assert args.ref is not None
        translate_to_file_with_beam_search(dest_fn, gpu, encdec_list, eos_idx, src_data,
                                           beam_width=beam_width,
                                           beam_pruning_margin=beam_pruning_margin,
                                           beam_score_coverage_penalty=beam_score_coverage_penalty,
                                           beam_score_coverage_penalty_strength=beam_score_coverage_penalty_strength,
                                           nb_steps=nb_steps,
                                           nb_steps_ratio=nb_steps_ratio,
                                           beam_score_length_normalization=beam_score_length_normalization,
                                           beam_score_length_normalization_strength=beam_score_length_normalization_strength,
                                           post_score_length_normalization=post_score_length_normalization,
                                           post_score_length_normalization_strength=post_score_length_normalization_strength,
                                           post_score_coverage_penalty=post_score_coverage_penalty,
                                           post_score_coverage_penalty_strength=post_score_coverage_penalty_strength,
                                           groundhog=groundhog,
                                           tgt_unk_id=tgt_unk_id,
                                           tgt_indexer=tgt_indexer,
                                           force_finish=force_finish,
                                           prob_space_combination=prob_space_combination,
                                           reverse_encdec=reverse_encdec,
                                           generate_attention_html=generate_attention_html,
                                           src_indexer=src_indexer,
                                           rich_output_filename=rich_output_filename,
                                           use_unfinished_translation_if_none_found=True)

        if ref is not None:
            bc = bleu_computer.get_bc_from_files(ref, dest_fn)
            print "bleu before unk replace:", bc
        else:
            print "bleu before unk replace: No Ref Provided"

        from nmt_chainer.utilities import replace_tgt_unk
        replace_tgt_unk.replace_unk(dest_fn, src_fn, dest_fn + ".unk_replaced", dic, remove_unk,
                                    normalize_unicode_unk,
                                    attempt_to_relocate_unk_source)

        if ref is not None:
            bc = bleu_computer.get_bc_from_files(ref, dest_fn + ".unk_replaced")
            print "bleu after unk replace:", bc
        else:
            print "bleu before unk replace: No Ref Provided"

    elif mode == "translate_attn":
        log.info("writing translation + attention as html to %s" % dest_fn)
        with cuda.get_device(gpu):
            assert len(encdec_list) == 1
            translations, attn_all = greedy_batch_translate(
                encdec_list[0], eos_idx, src_data, batch_size=mb_size, gpu=gpu,
                get_attention=True, nb_steps=nb_steps)
#         tgt_voc_with_unk = tgt_voc + ["#T_UNK#"]
#         src_voc_with_unk = src_voc + ["#S_UNK#"]
        assert len(translations) == len(src_data)
        assert len(attn_all) == len(src_data)
        attn_vis = AttentionVisualizer()
        for num_t in xrange(len(src_data)):
            src_idx_list = src_data[num_t]
            tgt_idx_list = translations[num_t][:-1]
            attn = attn_all[num_t]
#             assert len(attn) == len(tgt_idx_list)

            src_w = src_indexer.deconvert_swallow(src_idx_list, unk_tag="#S_UNK#") + ["SUM_ATTN"]
            tgt_w = tgt_indexer.deconvert_swallow(tgt_idx_list, unk_tag="#T_UNK#")
#             src_w = [src_voc_with_unk[idx] for idx in src_idx_list] + ["SUM_ATTN"]
#             tgt_w = [tgt_voc_with_unk[idx] for idx in tgt_idx_list]
#             for j in xrange(len(tgt_idx_list)):
#                 tgt_idx_list.append(tgt_voc_with_unk[t_and_attn[j][0]])
#
    #         print [src_voc_with_unk[idx] for idx in src_idx_list], tgt_idx_list

            attn_vis.add_plot(src_w, tgt_w, attn)

        attn_vis.make_plot(dest_fn)

    elif mode == "align":
        import nmt_chainer.utilities.visualisation as visualisation
        assert tgt_data is not None
        assert len(tgt_data) == len(src_data)
        log.info("writing alignment as html to %s" % dest_fn)
        with cuda.get_device(gpu):
            assert len(encdec_list) == 1
            loss, attn_all = batch_align(
                encdec_list[0], eos_idx, zip(src_data, tgt_data), batch_size=mb_size, gpu=gpu)
#         tgt_voc_with_unk = tgt_voc + ["#T_UNK#"]
#         src_voc_with_unk = src_voc + ["#S_UNK#"]

        assert len(attn_all) == len(src_data)
        plots_list = []
        for num_t in xrange(len(src_data)):
            src_idx_list = src_data[num_t]
            tgt_idx_list = tgt_data[num_t]
            attn = attn_all[num_t]
#             assert len(attn) == len(tgt_idx_list)

            alignment = np.zeros((len(src_idx_list) + 1, len(tgt_idx_list)))
            sum_al = [0] * len(tgt_idx_list)
            for i in xrange(len(src_idx_list)):
                for j in xrange(len(tgt_idx_list)):
                    alignment[i, j] = attn[j][i]
                    sum_al[j] += alignment[i, j]
            for j in xrange(len(tgt_idx_list)):
                alignment[len(src_idx_list), j] = sum_al[j]

            src_w = src_indexer.deconvert_swallow(src_idx_list, unk_tag="#S_UNK#") + ["SUM_ATTN"]
            tgt_w = tgt_indexer.deconvert_swallow(tgt_idx_list, unk_tag="#T_UNK#")
#             src_w = [src_voc_with_unk[idx] for idx in src_idx_list] + ["SUM_ATTN"]
#             tgt_w = [tgt_voc_with_unk[idx] for idx in tgt_idx_list]
#             for j in xrange(len(tgt_idx_list)):
#                 tgt_idx_list.append(tgt_voc_with_unk[t_and_attn[j][0]])
#
    #         print [src_voc_with_unk[idx] for idx in src_idx_list], tgt_idx_list
            p1 = visualisation.make_alignment_figure(
                src_w, tgt_w, alignment)
#             p2 = visualisation.make_alignment_figure(
#                             [src_voc_with_unk[idx] for idx in src_idx_list], tgt_idx_list, alignment)
            plots_list.append(p1)
        p_all = visualisation.Column(*plots_list)
        visualisation.output_file(dest_fn)
        visualisation.show(p_all)
#     for t in translations_with_attn:
#         for x, attn in t:
#             print x, attn


#         out.write(convert_idx_to_string([x for x, attn in t], tgt_voc + ["#T_UNK#"]) + "\n")

    elif mode == "score_nbest":
        log.info("opening nbest file %s" % nbest_to_rescore)
        nbest_f = codecs.open(nbest_to_rescore, encoding="utf8")
        nbest_list = [[]]
        for line in nbest_f:
            line = line.strip().split("|||")
            num_src = int(line[0].strip())
            if num_src >= len(nbest_list):
                assert num_src == len(nbest_list)
                if max_nb_ex is not None and num_src >= max_nb_ex:
                    break
                nbest_list.append([])
            else:
                assert num_src == len(nbest_list) - 1
            sentence = line[1].strip()
            nbest_list[-1].append(sentence.split(" "))

        log.info("found nbest lists for %i source sentences" % len(nbest_list))
        nbest_converted, make_data_infos = make_data.build_dataset_for_nbest_list_scoring(tgt_indexer, nbest_list)
        log.info("total %i sentences loaded" % make_data_infos.nb_ex)
        log.info("#tokens src: %i   of which %i (%f%%) are unknown" % (make_data_infos.total_token,
                                                                       make_data_infos.total_count_unk,
                                                                       float(make_data_infos.total_count_unk * 100) /
                                                                       make_data_infos.total_token))
        if len(nbest_list) != len(src_data[:max_nb_ex]):
            log.warn("mismatch in lengths nbest vs src : %i != %i" % (len(nbest_list), len(src_data[:max_nb_ex])))
            assert len(nbest_list) == len(src_data[:max_nb_ex])

        log.info("starting scoring")
        from nmt_chainer.utilities import utils
        res = []
        for num in xrange(len(nbest_converted)):
            if num % 200 == 0:
                print >>sys.stderr, num,
            elif num % 50 == 0:
                print >>sys.stderr, "*",

            res.append([])
            src, tgt_list = src_data[num], nbest_converted[num]
            src_batch, src_mask = utils.make_batch_src([src], gpu=gpu, volatile="on")

            assert len(encdec_list) == 1
            scorer = encdec_list[0].nbest_scorer(src_batch, src_mask)

            nb_batches = (len(tgt_list) + mb_size - 1) / mb_size
            for num_batch in xrange(nb_batches):
                tgt_batch, arg_sort = utils.make_batch_tgt(tgt_list[num_batch * nb_batches: (num_batch + 1) * nb_batches],
                                                           eos_idx=eos_idx, gpu=gpu, volatile="on", need_arg_sort=True)
                scores, attn = scorer(tgt_batch)
                scores, _ = scores
                scores = scores.data

                assert len(arg_sort) == len(scores)
                de_sorted_scores = [None] * len(scores)
                for xpos in xrange(len(arg_sort)):
                    original_pos = arg_sort[xpos]
                    de_sorted_scores[original_pos] = scores[xpos]
                res[-1] += de_sorted_scores
        print >>sys.stderr
        log.info("writing scores to %s" % dest_fn)
        out = codecs.open(dest_fn, "w", encoding="utf8")
        for num in xrange(len(res)):
            for score in res[num]:
                out.write("%i %f\n" % (num, score))



from collections import OrderedDict
import nmt_chainer.utilities.argument_parsing_tools

cell_description_keywords = OrderedDict([
    ("cell_type", str),
    ("sub_cell_type", str),
    ("nb_stacks", int),
    ("dropout", float),
    ("residual_connection", int),
    ("no_dropout_on_input", int),
    ("no_residual_connection_on_output", int),
    ("no_residual_connection_on_input", int),
    ("init_type", str),
    ("inner_init_type", str),
    ("lateral_init_type", str),
    ("upward_init_type", str),
    ("bias_init_type", str),
    ("forget_bias_init_type", str),
    ("init_scale", float),
    ("inner_init_scale", float),
    ("lateral_init_scale", float),
    ("upward_init_scale", float),
    ("bias_init_scale", float),
    ("forget_bias_init_scale", float),
    ("init_fillvalue", float),
    ("inner_init_fillvalue", float),
    ("lateral_init_fillvalue", float),
    ("upward_init_fillvalue", float),
    ("bias_init_fillvalue", float),
    ("forget_bias_init_fillvalue", float)
])


def create_cell_config_from_string(model_str):
    components = model_str.split(",")
    if ":" not in components[0]:
        components = ["cell_type:%s" % components[0]] + components[1:]
    keywords = {}
    for comp in components:
        assert ":" in comp
        key, val = comp.split(":")
        if key in cell_description_keywords:
            keywords[key] = cell_description_keywords[key](val)
        else:
            raise ValueError("bad cell parameter: %s (possible parameters: %s)" %
                             (comp, " ".join(cell_description_keywords.keys())))

    ordered_keywords = OrderedDict()
    for key in cell_description_keywords.keys():
        if key in keywords:
            ordered_keywords[key] = keywords[key]
    nmt_chainer.utilities.argument_parsing_tools.OrderedNamespace.convert_to_ordered_namespace(
        ordered_keywords)
#     ordered_keywords.set_readonly()
    return ordered_keywords
